/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "config.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <getopt.h>
#ifdef HAVE_LIBNCURSES
#include <curses.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include "version.h"
#include "main.h"
#include "error.h"
#include <sys/ioctl.h>

/************ Prototypes ************/

void main_trace_dumb(int n);
void main_trace_quiet(int);
void main_trace_curses(int n);
void main_finish_dumb();
void main_finish_quiet();
void main_finish_curses();
void main_init_curses();
void main_init_dumb();
void main_init_quiet();
void main_update_curses();
void main_update_dumb();
void main_update_quiet();
void main_do_options(int argc, char *argv[], char *envp[]);
void info();

/************************************/

char error_progname[PATH_MAX];
int main_paused = 0;

void (*main_finish_func)() = NULL;

void main_trace_dumb(int n)
{
	static int oldsec, oldusec, isconv;
	static double old2fps, oldfps, fps;
	static struct timeval after;
	static char src[PATH_MAX], dest[PATH_MAX], conv[PATH_MAX];
	int nr;

	if (n == 0) {
		gettimeofday(&after, 0);
		fps = oldfps = 0;
		isconv = 0;
		return;
	}
	nr = libgrab_getprogress(src,dest,conv);
	if (nr < 0) exit(0);
	if (main_paused) return;
	if (isconv) {
		if (!*src && !*dest && !*conv) return;
		printf("%s\n",conv);
	} else {
		if (!*src && !*dest && !*conv) {
			isconv = 1;
			return;
		}
		oldsec = after.tv_sec;
		oldusec = after.tv_usec;
		old2fps = oldfps;
		oldfps = fps;
		gettimeofday(&after, 0);
		fps = (1/((after.tv_sec + after.tv_usec/1.0e6) - (oldsec + oldusec/1.0e6)))+0.05;
		printf("%s (%2.01f fps)\n",dest,(fps+oldfps+old2fps)/3);
	}
}

void main_trace_quiet(int n)
{
	static char src[PATH_MAX], dest[PATH_MAX], conv[PATH_MAX];
	if (n == 0) return;
	if (libgrab_getprogress(src,dest,conv) < 0) exit(0);
}

#ifdef HAVE_LIBNCURSES
void main_trace_curses(int n)
{
	static int oldsec, oldusec, start, isconv;
	static double old2fps, oldfps, fps;
	static char src[PATH_MAX], dest[PATH_MAX], conv[PATH_MAX];
	static struct timeval after;
	int nr;

	if (n == 0) {
		gettimeofday(&after, 0);
		start = after.tv_sec;
		fps = oldfps = 0;
		isconv = 0;
		return;
	}
	nr = libgrab_getprogress(src,dest,conv);
	if (main_paused) return;
	if (nr < 0) exit(0);
	if (isconv) {
		if (!*src && !*dest && !*conv) return;
		attron(A_BOLD);
		mvprintw(17,17,conv);
		attroff(A_BOLD);
	} else {
		int i = nr-1;
		char *p = NULL;
		if (!*src && !*dest && !*conv) {
			isconv = 1;
			return;
		}
		p = strrchr(dest,'[');
		if (p) sscanf(p,"[%d",&i);
		attron(A_BOLD);
		mvprintw(17,17,dest);
		attroff(A_BOLD);
		oldsec = after.tv_sec;
		oldusec = after.tv_usec;
		old2fps = oldfps;
		oldfps = fps;
		gettimeofday(&after, 0);
		fps = (1/((after.tv_sec + after.tv_usec/1.0e6) - (oldsec + oldusec/1.0e6)))+0.05;
		mvprintw(6, 14, "%2.01f (%2.01f)   ", (fps+oldfps+old2fps)/3,after.tv_sec==start?0:(double)nr/(double)(after.tv_sec-start));
		mvprintw(5, 45, "%i:%i.%i ",(after.tv_sec-start)/3600,
			 ((after.tv_sec-start)/60)%60,(after.tv_sec-start)%60);
		mvprintw(5, 14, "%i  ",nr);
		mvprintw(6, 45, "%i  ", i-nr);
	}

	refresh();
}

void main_finish_curses()
{
	flushinp();
	endwin();
}

void main_init_curses()
{
	char tmp1[PATH_MAX], tmp2[PATH_MAX];

	initscr(); cbreak(); noecho();		// init ncurses
	leaveok(stdscr, TRUE);			// disable cursor
	wresize(stdscr,24,80);			// resize win to 80x24
	refresh();				// flush the screen 
	start_color();				// set collors
	init_pair(1,COLOR_WHITE,COLOR_BLUE);
	init_pair(2,COLOR_RED,COLOR_BLUE);
	init_pair(3,COLOR_CYAN,COLOR_BLUE);
	wbkgd(stdscr, COLOR_PAIR(1));
	wattrset(stdscr, COLOR_PAIR(1));
	attron(A_BOLD);
	newwin(0,0,0,0);
	box(stdscr, 0, 0);			// 80x24 box
	
	mvprintw(14, 2, " Keys: ");

	libgrab_inquire("width",tmp1);
	libgrab_inquire("height",tmp2);
	mvprintw(7,14," %sx%s",tmp1,tmp2);
	
	attroff(A_BOLD);
	mvprintw(14,10,"Q	- Quit");
	mvprintw(15,10,"P	- Pause");
	mvprintw(16,10,"S	- Sync HDDs");

	mvprintw(5, 2, " Image Nr.:");
	mvprintw(7, 2, "Image size:");
	mvprintw(6, 2, "       FPS:");
	mvprintw(9, 2, "    Status:");

	mvprintw(5, 32,"       Time:");
	mvprintw(6, 32,"Frames lost:");
	

	mvprintw(17, 2,"Grabbing into:");

	attron(A_BOLD);
	mvprintw(9, 14, "Working...");		// status

	wattrset(stdscr, COLOR_PAIR(2));
	mvprintw(3, 1, "------------------------------------------------------------------------------");
	mvprintw(11, 1, "------------------------------------------------------------------------------");
	wattrset(stdscr, COLOR_PAIR(3));
	attron(A_BOLD);
	mvprintw(0,2,"< bttvgrab " VERSION " (v4l) by J. Walter/A. Kopacz >");
	mvprintw(23,30,"< http://moes.pmnet.uni-oldenburg.de/bttvgrab/ >");
	wattrset(stdscr, COLOR_PAIR(1));
	
	refresh();
	main_finish_func = main_finish_curses;
}

void main_update_curses()
{
	switch(getch()) { 
	case 'p': case 'P':
		attron(A_BOLD);
		if (!main_paused) {
			mvprintw(9,14,"PAUSED - Push any key (except s/q) to continue");
			libgrab_control(C_PAUSE);
			main_paused = 1;
		} else {
			main_trace_curses(0);
			mvprintw(9,14,"Working...                                    ");
			libgrab_control(C_CONT);
			main_paused = 0;
		}
		refresh();
		break;
	case 's': case 'S':
		attron(A_BOLD);
		mvprintw(10,14,"Syncing HDDs...");
		refresh();	      
		sync();
		attron(A_BOLD);
		mvprintw(10,14,"               ");
		refresh();
		break;
	case 'q': case 'Q':
		libgrab_control(C_STOP);
		exit(0);
		
	default: 
		break;
	}
}

void (*main_init_func)() = main_init_curses;
void (*main_update)() = main_update_curses;
void (*main_trace_func)(int) = main_trace_curses;
#else
void (*main_init_func)() = main_init_dumb;
void (*main_update)() = main_update_dumb;
void (*main_trace_func)(int) = main_trace_dumb;
#endif

void main_finish_dumb()
{
	printf("done.\n");
}

void main_finish_quiet()
{
}

void main_finish()
{
	if (main_finish_func) (*main_finish_func)();
	main_trace_func = NULL;
	main_update = NULL;
	if (*error_text) fprintf(stderr,"%s\n",error_text);
}

void main_init_dumb()
{
	setbuf(stdout,NULL);
	main_finish_func = main_finish_dumb;
}

void main_init_quiet()
{
	setbuf(stdout,NULL);
	setbuf(stdin,NULL);
	main_finish_func = main_finish_quiet;
}

void main_update_dumb()
{
	switch(getc(stdin)) { 
	case '\n':
		if (main_paused) {
			main_paused = 0;
			libgrab_control(C_CONT);
		} else {
			main_paused = 1;
			libgrab_control(C_PAUSE);
			printf("paused\n");
		}
		break;
	case 'q':
		libgrab_control(C_STOP);
		exit(0);
	default:
		break;
	}
}

void main_update_quiet()
{
}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////                                             ///////////////
/////////////                 Main loop                   ///////////////
/////////////                                             ///////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[], char *envp[])
{
	fd_set fds;
	FD_ZERO(&fds);

	strcpy(error_progname,argv[0]);

	libgrab_init();

	main_do_options(argc, argv, envp);

#if defined(VISI) && VISI > 1
#include "../visi-extra.inc"
#endif

	(*main_init_func)();
	(*main_trace_func)(0);
	atexit(main_finish);
	libgrab_control(C_START);
loop:
	if (main_trace_func != main_trace_quiet) FD_SET(0,&fds);
	FD_SET(libgrab_progresspipe,&fds);
	FD_SET(libgrab_errorpipe,&fds);
	select(libgrab_errorpipe+libgrab_progresspipe+1,&fds,NULL,NULL,NULL);

	if (FD_ISSET(libgrab_errorpipe,&fds)) exit(1);
	if (FD_ISSET(libgrab_progresspipe,&fds) && main_trace_func) (*main_trace_func)(1);
	if (FD_ISSET(0,&fds) && main_update) (*main_update)();
	goto loop;
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

#define OPTSET(c,opt,val) case c: libgrab_configure(#opt,val); break;
#define OPT(c,opt) OPTSET(c,opt,optarg);

void main_do_options(int argc, char *argv[], char *envp[])
{
	int optchar, height = 0;
	char tmp[PATH_MAX], *p;
	libgrab_configure("width","640");

	while ((optchar = getopt(argc, argv, "O:b:N:z:Z:G:S:k:W:w:p:s:l:f:F:P:o:g:d:tihQy")) != -1) {
		switch (optchar) {
		OPT('z',rate);
		OPT('Z',srate);
		OPT('w',width);
		case 'W':
			height = 1;
		OPT(1,height);
		OPT('G',grabdevice);
		OPT('k',skip);
		OPT('b',fbs);
		OPT('N',norm);
		OPT('S',channel);
		OPT('p',output_data);
		OPT('s',sleeptime);
		OPT('l',limit);
		OPT('o',grab_file_fmt);
		OPT('g',grablog);
		OPT('O',output_sound_fmt);
		OPT('F',osoundname);
		OPT('P',isounddata);
		case 'f':
			p = strchr(optarg,':');
			if (p) {
				*p = 0;
				libgrab_configure("grab_transport",optarg);
				*p = ':';
				libgrab_configure("grabname",p+1);
			} else libgrab_configure("grabname",optarg);
			break;
		OPTSET('t',do_trace,"1");
		OPTSET('y',do_sync,"1");
		OPTSET('Q',do_sound,"0");
		case 'i': info();
		case 'd':
			ASSERT(strlen(optarg) == 1,(ARGPMSG("display type must be one of c,d,q,t")));
			switch (*optarg) {
			case 't': {
				char *command;
				int i, len = 0;
				for (i = 1; i < argc; i++) len += strlen(argv[i]) + 1;
				len += strlen(WISH " " DATADIR "/bttvgrab.tcl -- ");
				command = malloc(len);
				strcpy(command,WISH " " DATADIR "/bttvgrab.tcl --");
				for (i = 1; i < argc; i++) {
					strcat(command," ");
					strcat(command,argv[i]);
				}
				NOTNEG1(execle("/bin/sh","/bin/sh","-c",command,NULL,envp),(ERRMSG("executing "WISH)));
				ERROR(NOERR("done."));
				exit(0);
			}
			case 'c':
#ifdef HAVE_LIBNCURSES
				main_init_func = main_init_curses;
				main_trace_func = main_trace_curses;
				main_update = main_update_curses;
#else
				ASSERT(0,(MSG("curses display not compiled in")));
#endif
				break;
			case 'd':
				main_init_func = main_init_dumb;
				main_trace_func = main_trace_dumb;
				main_update = main_update_dumb;
				break;
			case 'q':
				main_init_func = main_init_quiet;
				main_trace_func = main_trace_quiet;
				main_update = main_update_quiet;
				break;
			default:
				ASSERT(0,(ARGPMSG("display type must be one of c,d,q,t")));
				break;
			}
			break;
		case 'h': 
		default:
			printf("bttvgrab " VERSION "\n");
			printf("(c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>\n");
			printf("Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/\n\n");

			printf("Options:\n");
			printf("-w [32-768]      : width, default = 640\n");
			printf("-W [32-576]      : height, default = width*3/4\n");
			printf("-o [format]      : output file format, default = last, available:\n");
			libgrab_inquire("grab_file_fmt list",tmp);
			printf("%s\n",tmp);
			printf("-t               : trace lost frames by leaving out filenumbers\n");
			printf("-g [filename]    : create a grab logfile, default = disabled\n");
			printf("-b [2-32]        : use at most this many grab buffers, default = driver max.\n");
			printf("-G [devicename]  : grab device, default = /dev/bttv\n");
			printf("-S [channelnr]   : source channel number, default=0\n");
			printf("-N [normname]    : video norm, available: PAL, NTSC, SECAM, AUTO (default)\n");
			printf("-p [name]=[val]  : extended output parameters, depends on output format\n");
			printf("                   try quality=n, many more available\n");
			printf("-s [1-?]         : sleeptime, default = 1\n");
			printf("                   sleeping time in seconds\n");
			printf("-k [n]           : frame skip interval, grab a frame after <n> frames\n");
			printf("                   skipped, or skip a frame after -<n> frames grabbed\n");
			printf("-l [1-?]         : limit, exit after <limit> pictures.\n");
			printf("-f [URL]         : filename/URL for output-picture(s)\n");
			printf("                   if it ends with '/', a default filename is appended\n");
			printf("                   otherwise use %%04d (or similar) for numbering\n");
			printf("                   currently supported methods: file: (default),\n");
			printf("                   net: (use server address as name), and\n");
			printf("                   webcam: (like file:, plus some help files)\n");
			printf("-F [filename]    : filename for output wav, default is ./grab.wav\n");
			printf("                   if it ends with '/', a default filename is appended\n");
			printf("-O [format]      : output sound format, default = first, available:\n");
			libgrab_inquire("output_sound_fmt list",tmp);
			printf("%s\n",tmp);
			printf("-P [opt=value]   : sound parameters, opt depends on format, try device,\n");
			printf("                   rate, bits, channels\n");
			printf("-Q               : quiet, do not record sound.\n");
			printf("-d [c|d|q|t]     : display type (curses/dumb/quiet/TclTk), default = curses\n");
			printf("-y               : synchronous output, use this if you have large\n");
			printf("                 : video sequences to grab, needs *fast* hard disk\n");
			printf("-i               : shows some v4l informations.\n");
			printf("-h               : this help screen.\n\n");
			exit(0);
		}
	}
	if (!height) {
		libgrab_inquire("width",tmp);
		sprintf(tmp,"%i",atoi(tmp)*3/4);
		libgrab_configure("height",tmp);
	}
}

void info()
{
	char tmp[4096];
	printf("bttvgrab "    VERSION    " - Video4Linux Informations\n");
	printf("*****************************************************\n\n");
	libgrab_inquire("grabinfo",tmp);
	puts(tmp);
	printf("\n");
	exit(0);
}
