/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#ifdef HAVE_LINUX_VIDEODEV_H
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <pthread.h>
#ifdef HAVE_SEMAPHORE_H
#include <semaphore.h>
#endif
#include "grab.h"
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) grab_ ## x
#ifdef __CYGWIN__
#define PTH_T(x) &x
#else
#define PTH_T(x) x
#endif

char grab_device[PATH_MAX]="/dev/video";
int grab_width = 640;
int grab_height = 480;
int grab_format = VIDEO_PALETTE_RGB24;
char grab_log[PATH_MAX] = "";
int grab_sleep = 0;

int grab_init_p = 0;
int grab_fbttv = 0, grab_fbs = 32, grab_channels = 0;
unsigned char **grab_buf = NULL;
int *grab_buf_picnr, *grab_bufst, *grab_bufqueue;
int grab_picnr = 0;
pid_t grab_child = 0;
int grab_norm = 3;
FILE *grab_log_fp = NULL;
int grab_skip = 0, grab_skip_count = 0;
pthread_t grab_thread = 0;
struct video_mbuf grab_mbuf;
struct video_channel grab_channel;
struct video_audio grab_audio;
pthread_mutex_t grab_stmutex;
sem_t grab_newframe;
sem_t grab_framedone;

#define ST_EMPTY 0
#define ST_ARMED 1
#define ST_READY 2
#define ST_INUSE 3

/************ Prototypes ************/

void grab_init();
void grab_start();
void grab_info(char *bptr);
void *grab_child_start(void *ign);
int grab_get();
int grab_get_internal();
int grab_get_picnr();
void grab_release();
void grab_stop();
void grab_end();

/************************************/

#define GRAB_NORMS 4
const char *grab_norms[GRAB_NORMS] = {"PAL","NTSC","SECAM","AUTO"};
const int grab_normvals[GRAB_NORMS] = {VIDEO_MODE_PAL,VIDEO_MODE_NTSC,VIDEO_MODE_SECAM,VIDEO_MODE_AUTO};

#define GRAB_COLORS 8
const char *grab_color_names[GRAB_COLORS] = {"RGB24","RGB16","PAL8","YUV422","YUV420", "GREY8", "RGB32", "RGB15"};
const char *grab_color_subst[GRAB_COLORS] = {"RGB24S","RGB16","PAL8","YUV422S","YUV422S", "GREY8", "RGB32", "RGB15"};
const int grab_color_values[GRAB_COLORS] = {VIDEO_PALETTE_RGB24,
					    VIDEO_PALETTE_RGB565,
					    VIDEO_PALETTE_HI240,
					    VIDEO_PALETTE_YUV422,
					    VIDEO_PALETTE_YUV422,
					    VIDEO_PALETTE_GREY,
					    VIDEO_PALETTE_RGB32,
					    VIDEO_PALETTE_RGB555};

START_OPTIONS
INTOPTION(sleep,0,1);
STROPTION(device); if (set) {
	struct video_capability vc;
	int fh;
	if ((fh = open(grab_device,O_RDONLY)) == -1) {
		grab_channels = 3;
		grab_channel.channel = 0;
		grab_channel.norm = grab_normvals[3];
	} else {
		NOTNEG1(ioctl(fh,VIDIOCGCAP,&vc),(ERRMSG("VIDIOCGCAP")));
		grab_channels = vc.channels;
		grab_channel.channel = 0;
		grab_channel.norm = grab_normvals[3];
		SAFE_CLOSE(fh,grab_device);
	}
	grab_audio.audio = 0;
}
INTINFO(channels);
OPTION(channel list) NOSET O_GET {
	struct video_channel vch;
	int fh;
	if (grab_channels == 0) grab_configure("device",grab_device);
	if ((fh = open(grab_device,O_RDONLY)) == -1) {
		strcpy(value,"unknown");
	} else {
		*value = 0;
		for (vch.channel = 0; vch.channel < grab_channels; vch.channel++) {
			NOTNEG1(ioctl(fh,VIDIOCGCHAN,&vch),(ERRMSG("VIDIOCGCHAN")));
			strcat(value,vch.name);
			strcat(value," ");
		}
		if (!(*value)) value[strlen(value)-1] = 0;
		SAFE_CLOSE(fh,grab_device);
	}
}
OPTION(channel) O_SET {
	int fh;
	int ch = atoi(value);
	if (grab_channels == 0) grab_configure("device",grab_device);
	ASSERT(ch >= 0 && ch < grab_channels,(INTMSG("channel out of range")));
	grab_channel.channel = ch;
} O_GET {
	sprintf(value,"%i",grab_channel.channel);
}
OPTION(info) NOSET O_GET grab_info(value);
OPTION(norm) O_SET {
	int fh; int i;
	for (i = 0; i < GRAB_NORMS; i++) {
		if (!strcmp(value,grab_norms[i])) {
			break;
		}
	}
	ASSERT(i < GRAB_NORMS,(MSG("Unknown video norm '%s'"),value));
	grab_channel.norm = grab_normvals[i];
	grab_norm = i;
	if ((fh = open(grab_device,O_RDONLY)) != -1) {
		ioctl(fh,VIDIOCSCHAN,&grab_channel);
		SAFE_CLOSE(fh,grab_device);
	}
} O_GET {
	strcpy(value,grab_norms[grab_norm]);
}
INTOPTION(width,32,768);
INTOPTION(height,32,576);
INTOPTION(format,0,INT_MAX);
OPTION(color_fmt) O_SET {
	int i;
	for (i = 0; i < GRAB_COLORS; i++) {
		if (!strcmp(value,grab_color_names[i])) {
			grab_format = grab_color_values[i];
			break;
		}
	}
	ASSERT(i < GRAB_COLORS,(INTMSG("unknown color format '%s'"),value));
} O_GET {
	int i;
	for (i = 0; i < GRAB_COLORS; i++) {
		if (grab_format == grab_color_values[i]) {
			strcpy(value,grab_color_subst[i]);
			break;
		}
	}
	ASSERT(i < GRAB_COLORS,(INTMSG("unknown color format '%i'"),grab_format));
}
INTOPTION(fbs,2,32);
PTRINFO(buf,grab_fbs);
STROPTION(log);
INTOPTION(skip,-100,100);
END_OPTIONS

void grab_dummy(int fbnr, int picnr) {}

 void (*grab_signal)(int fbnr, int picnr) = grab_dummy;

void grab_init()
{
	if (grab_init_p) return;
	grab_init_p = 1;

	memset(&grab_channel,0,sizeof(grab_channel));
	memset(&grab_audio,0,sizeof(grab_audio));

	grab_channel.norm=grab_normvals[3];
	grab_channels = 3;

	EQNULL((void *)sem_init(&grab_newframe,0,0),(MSG("sem_init")));
	EQNULL((void *)sem_init(&grab_framedone,0,0),(MSG("sem_init")));
	EQNULL((void *)pthread_mutex_init(&grab_stmutex,NULL),(MSG("pthread_mutex_init")));
}

void grab_start()
{
	struct video_capability vcapability;
	pthread_attr_t pattr;
	int i;

	if (grab_fbttv) return;

	SAFE_OPEN(grab_fbttv,grab_device,O_RDWR);
	grab_audio.audio = 0;

	ioctl(grab_fbttv,VIDIOCGAUDIO,&grab_audio);
	grab_audio.volume=32767;
	grab_audio.flags &= ~VIDEO_AUDIO_MUTE;
	ioctl(grab_fbttv,VIDIOCSAUDIO,&grab_audio);

	grab_channel.norm=grab_normvals[grab_norm];
	ioctl(grab_fbttv,VIDIOCSCHAN,&grab_channel);

	NOTNEG1(ioctl(grab_fbttv,VIDIOCGMBUF,&grab_mbuf),(ERRMSG("VIDIOCGMBUF")));
	if (grab_mbuf.frames < grab_fbs) grab_fbs = grab_mbuf.frames;

	SAFE_MALLOC(grab_bufqueue,grab_fbs*sizeof(int));
	SAFE_MALLOC(grab_bufst,grab_fbs*sizeof(int));
	SAFE_MALLOC(grab_buf_picnr,grab_fbs*sizeof(int));
	for (i = 0; i < grab_fbs; i++) {
		grab_bufqueue[i] = -1;
		grab_bufst[i] = ST_EMPTY;
		grab_buf_picnr[i] = -1;
	}

	NOTNEG1(ioctl(grab_fbttv,VIDIOCGCAP,&vcapability),(ERRMSG("VIDIOCGCAP")));

	RANGE(grab_width,vcapability.minwidth,vcapability.maxwidth,"grab width",MSG);
	RANGE(grab_height,vcapability.minheight,vcapability.maxheight,"grab height",MSG);

	SAFE_MALLOC(grab_buf,grab_fbs*sizeof(char *));
	grab_buf[0] = (void *)-1;
	NOTNEG1((int)(grab_buf[0]=mmap(0,(grab_mbuf.size/grab_mbuf.frames)*grab_fbs,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fbttv,0)),
		(ERRMSG("mmap %i bytes"),(grab_mbuf.size/grab_mbuf.frames)*grab_fbs));

	for (i = 1; i < grab_fbs; i++) {
		grab_buf[i] = grab_buf[i-1]+(grab_mbuf.size/grab_mbuf.frames);
	}

	while (!sem_trywait(&grab_newframe));
	while (!sem_trywait(&grab_framedone));
	// start grab thread
	pthread_attr_init(&pattr);
	pthread_attr_setdetachstate(&pattr,PTHREAD_CREATE_DETACHED);
	EQNULL((void *)pthread_create(&grab_thread,&pattr,grab_child_start,NULL),(ERRMSG("pthread_create")));
	pthread_attr_destroy(&pattr);

	sem_wait(&grab_newframe);
}

void grab_info(char *bptr)
{
	struct video_capability vcapability;
	struct video_tuner vtuner;
	struct video_buffer vbuffer;
	struct video_channel vchan;
	struct video_audio vaudio;
	int i;
	char *device_cap[] = {
		"capture", "tuner", "teletext", "overlay", "chromakey", "clipping",
		"frameram", "scales", "monochrome", "subcapture", NULL };

	SAFE_OPEN(grab_fbttv,grab_device,O_RDWR);

	NOTNEG1(ioctl(grab_fbttv,VIDIOCGCAP,&vcapability),(ERRMSG("VIDIOCGCAP")));
	
	NOTNEG1(ioctl(grab_fbttv,VIDIOCGFBUF,&vbuffer),(ERRMSG("VIDIOCGFBUF")));
	
	sprintf(bptr,"card_name     : %s\n",vcapability.name);
	bptr += strlen(bptr);
	sprintf(bptr,"flags         :");
	bptr += strlen(bptr);
	for (i = 0; device_cap[i] != NULL; i++) {
		if (vcapability.type & (1 << i)) {
			sprintf(bptr," %s",device_cap[i]);
			bptr += strlen(bptr);
		}
	}
	
	sprintf(bptr,"\nchannels      : %d (",vcapability.channels);
	bptr += strlen(bptr);
	for (i = 0; i < vcapability.channels; i++) {
		vchan.channel = i;
		if (ioctl(grab_fbttv,VIDIOCGCHAN,&vchan) != -1) {
			sprintf(bptr," %s",vchan.name);
			bptr += strlen(bptr);
		}
	}
	sprintf(bptr," )\n");
	bptr += strlen(bptr);

	sprintf(bptr,"audio devices : %d (",vcapability.audios);
	bptr += strlen(bptr);
	for (i = 0; i < vcapability.audios; i++) {
		vaudio.audio = i;
		if (ioctl(grab_fbttv,VIDIOCGAUDIO,&vaudio) != -1) {
			sprintf(bptr," %s",vaudio.name);
			bptr += strlen(bptr);
		}
	}
	sprintf(bptr," )\n");
	bptr += strlen(bptr);

	sprintf(bptr,"size (max/min): %d*%d/%d*%d\n",vcapability.maxwidth,vcapability.maxheight,vcapability.minwidth,vcapability.minheight);
	bptr += strlen(bptr);

	vtuner.tuner = 0;
	if (ioctl(grab_fbttv,VIDIOCGTUNER,&vtuner) != -1) {
		sprintf(bptr,"tuner         : %s (Freq.: %lu-%lu)\n",vtuner.name,vtuner.rangelow,vtuner.rangehigh);
		bptr += strlen(bptr);
	}

	sprintf(bptr,"grab_buffer   : base=%p size=%dx%d depth=%d bpl=%d\n",vbuffer.base,vbuffer.width,vbuffer.height,vbuffer.depth,vbuffer.bytesperline);
}

#define iioctl(a,b,c) NOTNEG1(ioctl(a,b,c),(ERRMSG(#b)))

void *grab_child_start(void *ign)
{
	struct video_mmap *gb;
	struct timeval grabtime;
	int nextnum, newnum, tmp, i, grab_bufnum;

	if (*grab_log) {
		SAFE_FOPEN(grab_log_fp,grab_log,"w");
	}

	pthread_mutex_lock(&grab_stmutex);
	SAFE_MALLOC(gb,grab_fbs*sizeof(struct video_mmap));

	for (i = 0; i < grab_fbs; i++) {
		gb[i].width = grab_width;
		gb[i].height = grab_height;
		gb[i].format = grab_format;
		gb[i].frame=i;
	}

	grab_bufnum = grab_fbs;
	grab_picnr = 0;
	/////////////////// grab loop ///////////////////////
	iioctl(grab_fbttv, VIDIOCMCAPTURE, &(gb[1]));
	iioctl(grab_fbttv, VIDIOCMCAPTURE, &(gb[0]));

	grab_bufst[0]++;
	grab_bufst[1]++;
	grab_buf_picnr[0] = grab_picnr++;
	grab_buf_picnr[1] = grab_picnr++;

	pthread_mutex_unlock(&grab_stmutex);


	nextnum = 0;
	newnum = 1;
	iioctl(grab_fbttv, VIDIOCSYNC, &newnum);
	(*grab_signal)(0,grab_picnr);
	sem_post(&grab_newframe);
	iioctl(grab_fbttv, VIDIOCMCAPTURE, &(gb[1]));
	if (grab_sleep || grab_bufnum == 2) {
		if (!grab_sleep) sem_post(&grab_framedone);
		goto loop2;
	}
loop:
	iioctl(grab_fbttv, VIDIOCSYNC, &nextnum);
//	(*grab_signal)(newnum,grab_buf_picnr[newnum]);

	pthread_mutex_lock(&grab_stmutex);

	grab_bufst[nextnum]++;
	for (i = 0; i < grab_bufnum; i++) if (grab_bufst[i] == ST_EMPTY) break;
	if (i == grab_bufnum) {
		tmp = nextnum;
		for (i = 0; i < grab_bufnum; i++) {
			if (grab_bufst[i] == ST_READY &&
				grab_buf_picnr[i] < grab_buf_picnr[tmp]) {
				tmp = i;
			}
		}
		for (i = 0; i < grab_bufnum; i++)
			if (grab_bufqueue[i] == tmp) break;
		while (i < grab_bufnum-1) {
			grab_bufqueue[i] = grab_bufqueue[i+1];
			i++;
		}
		for (i = 0; i < grab_bufnum; i++) if (grab_bufqueue[i] == -1) {
			grab_bufqueue[i] = nextnum;
			break;
		}
		nextnum = newnum;
		newnum = tmp;
		if (*grab_log) fprintf(grab_log_fp,"%i lost\n",grab_buf_picnr[newnum]);
	} else {
		if (*grab_log) {
			gettimeofday(&grabtime,0);
			fprintf(grab_log_fp,"%i %li %li\n",grab_buf_picnr[nextnum],grabtime.tv_sec,grabtime.tv_usec);
		}
		tmp = i;
		for (i = 0; i < grab_bufnum; i++) if (grab_bufqueue[i] == -1) {
			grab_bufqueue[i] = nextnum;
			break;
		}
		nextnum = newnum;
		newnum = tmp;
		sem_post(&grab_newframe);
	}
	grab_bufst[newnum] = ST_ARMED;
	grab_buf_picnr[newnum] = grab_picnr++;

	pthread_mutex_unlock(&grab_stmutex);

//	printf("-%i = %i\n",newnum,grab_picnr);
	iioctl(grab_fbttv, VIDIOCMCAPTURE, gb+newnum);

	goto loop;

// if we have only 2 grab_buffers, we have to wait for the main thread each frame
loop2:
	iioctl(grab_fbttv, VIDIOCSYNC, &nextnum);
//	(*grab_signal)(newnum,grab_buf_picnr[newnum]);

	if (!grab_sleep) sem_wait(&grab_framedone);
	pthread_mutex_lock(&grab_stmutex);
		
	grab_buf_picnr[nextnum] = grab_picnr++;
	grab_bufst[nextnum]++;
	grab_bufqueue[0] = nextnum;
	tmp = nextnum;
	nextnum = newnum;
	newnum = tmp;
	grab_bufst[newnum] = ST_ARMED;
	grab_buf_picnr[newnum] = grab_picnr++;
	sem_post(&grab_newframe);
	pthread_mutex_unlock(&grab_stmutex);

	if (grab_sleep) sem_wait(&grab_framedone);
	iioctl(grab_fbttv, VIDIOCMCAPTURE, gb+newnum);

	goto loop2;

}

int grab_get()
{
	if (grab_skip < 0) {
		grab_skip_count--;
		if (grab_skip_count == grab_skip) {
			grab_skip_count = 0;
			grab_get_internal();
			grab_release();
		}
	} else if (grab_skip > 0) {
		int i;
		for (i = 0; i < grab_skip; i++) {
			grab_get_internal();
			grab_release();
		}
	}
	return grab_get_internal();
}

int grab_get_internal()
{
	int x;
	sem_wait(&grab_newframe);
	pthread_mutex_lock(&grab_stmutex);

	x = grab_bufqueue[0];

	pthread_mutex_unlock(&grab_stmutex);
	return x;
}

int grab_get_picnr()
{
	return grab_buf_picnr[grab_bufqueue[0]];
}


void grab_release()
{
	int i;
	pthread_mutex_lock(&grab_stmutex);

	grab_bufst[grab_bufqueue[0]] = ST_EMPTY;
	for (i = 1; i < grab_fbs; i++) grab_bufqueue[i-1] = grab_bufqueue[i];
	grab_bufqueue[grab_fbs-1]  = -1;

	sem_post(&grab_framedone);
	pthread_mutex_unlock(&grab_stmutex);
}

void grab_stop()
{
	if (grab_thread) {
		pthread_t x = pthread_self();
		if (!pthread_equal(PTH_T(x),PTH_T(grab_thread))) pthread_kill(PTH_T(grab_thread),SIGABRT);
		grab_thread = 0;
	}
	if (grab_buf) {
		if (grab_buf[0] != (void *)-1) munmap(grab_buf[0],(grab_mbuf.size/grab_mbuf.frames)*grab_fbs);
		SAFE_FREE(grab_buf);
	}
	SAFE_FREE(grab_bufqueue);
	SAFE_FREE(grab_bufst);
	SAFE_FREE(grab_buf_picnr);
	if (grab_log_fp) {
		fclose(grab_log_fp);
		grab_log_fp = NULL;
	}
	if (grab_fbttv) {
		if (grab_fbttv != -1) close(grab_fbttv);
		grab_fbttv = 0;
	}
}

void grab_end()
{
	grab_stop();
}

#else
#warning no grab device available
#include <string.h>
#include "error.h"

 void (*grab_signal)(int fbnr, int picnr) = NULL;

void grab_init()
{
}

void grab_start()
{
	ERROR(MSG("grabbing not configured"));
	END(1);
}

int grab_get_picnr()
{
	ERROR(MSG("grabbing not configured"));
	END(1);
	return -1;
}

void grab_get(int n)
{
	ERROR(MSG("grabbing not configured"));
	END(1);
}

void grab_release(int n)
{
	ERROR(MSG("grabbing not configured"));
	END(1);
}

void grab_configure(const char *n, const char *v)
{
}

void grab_inquire(const char *n, char *v)
{
}

void grab_stop()
{
	ERROR(MSG("grabbing not configured"));
	END(1);
}

void grab_end()
{
}
#endif
