#ifndef COMMON_H
#define COMMON_H

#include "parm.h"
#include "qsize.h"

#define real float
#define RCOPY(X,Y,N) memcpy(Y, X,(N)*sizeof(real))
#define EPSILON 1.0e-35

/* Use hand-pipelined loops for higher speed on 21000 */

//#define PIPELINE 1

#define REALZEROP(x) ((x<EPSILON)&&(x>-EPSILON))

#define CLIPP(X,LOW,HIGH) ((X)<(LOW)?(LOW):(X)>(HIGH)?(HIGH):(X))

#define FLOAT2FIX(X,N) ((int)  (X))
#define FIX2FLOAT(X)   ((float)(X))
#define VOLATILE

#define sig_scale(SCALE,A,B) \
({int __i;for(__i=0;__i<IDIM;__i++)(B)[__i]=SCALE*(A)[__i];})

#define sub_sig(A,B,C) \
({int __i;for(__i=0;__i<IDIM;__i++) (C)[__i]=(A)[__i]-(B)[__i];})

#define pm
#define dm

/* Circular Buffer Register numbers for ADSP21000   */

#define NPUT 8
#define NGET 9

#define real float
#define RCOPY(X,Y,N) memcpy(Y, X,(N)*sizeof(real))
#define EPSILON 1.0e-35

#define IF(X) if(X)
#define ZARR(A) ({for (i=sizeof(A)/sizeof(A[0])-1; i>=0 ; i--) A[i] = 0.0;})
#if 0
#define CHECK1(V) ({if (! (V>=-10000. && V<=10000.)) abort();})
#define CHECK(A,N) ({int __i;for(__i=0;__i<N;__i++)CHECK1((A)[__i]);})
#else
#define CHECK1(V) 
#define CHECK(A,N) 
#endif

#define PWFILTER
#define noCBDEBUG

#define QMEM
#define LPC_MEM
#define PSF_LPC_MEM
#define WIN_MEM
#define CBMEM
#define COEFF_MEM 
#define REG(X)

/* Update obsoleted atomic array */

#define UPDATE(NAME) ({int _i;if(NAME ## _obsolete_p) \
		       for(_i=sizeof(NAME)/sizeof(NAME[0])-1;_i>=0;_i--)\
		       NAME[_i]=NAME ## _next[_i]; NAME ## _obsolete_p=0;})

/* Copy L words to X from circular buffer CIRC *ending* at offset EOS. 
   CL is the size of circular buffe CIRC */

#define CIRCOPY(X,CIRC,EOS,L,CL) ({int _i1,_i2,_i,_lx=0;\
 if((EOS)>=(L)){_i1=(EOS)-(L);_i2=(CL);} \
 else            {_i1=0;        _i2=(CL)+(EOS)-(L);}\
 for(_i=_i2; _i<(CL);   _i++)X[_lx++]=CIRC[_i];\
 for(_i=_i1; _i<(EOS); _i++)X[_lx++]=CIRC[_i];})

/* get queue index of the most recent vector */

#define QINDEX  ({int __qx=this->vector_end-this->thequeue; \
		      __qx ? QSIZE-IDIM : __qx-IDIM;})

#define STATE_MEM
#define ZF_MEM   
#define ZI_MEM   

#define SPORDER 10

#define DECIM 4
#define PMSIZE  (NPWSZ+KPMAX)	/* Postfilter Memory SIZE */
#define PDMSIZE (PMSIZE/DECIM)  /* Post. Decim. Memory SIZE */
#define DPERMAX (KPMAX/DECIM)	/* Max. Decimated Period */
#define DPERMIN (KPMIN/DECIM)	/* Min. Decimated Period */
#define SHIFTSZ (KPMAX+NPWSZ-NFRSZ+IDIM)

typedef struct {
	unsigned char buf;
	int bits;

	int d_vec_start;	/* Index of the start of the decoded speech vector */
	int d_vec_end;	/* Index of the end   of the decoded speech vector */
	int w_vec_start;	/* Index of the start of vector being written */
	int w_vec_end;	/* Index of the end   of vector being written */

	int dec_end;	/* Index of the end of the decoded speech */
	real QMEM thequeue[QSIZE];
	real QMEM * VOLATILE vector_end;

	real pwf_z_vec[LPCW+1], pwf_p_vec[LPCW+1], pwf_old_input[LPCW+NFRSZ+NONRW], pwf_rec[LPCW+1];

	real facv[LPC+1];
	real bsf_old_input[LPC+NFRSZ+NONR], bsf_rec[LPC+1]; 

	real g_rec[LPCLG+1];	/* Recursive part for Hybrid Window */
	real g_old_input[LPCLG + NUPDATE + NONRLG];

	real a10[11];
	real k10;

	real firmem[LPCW+IDIM];
	real iirmem[LPCW+IDIM];

	real STATE_MEM statelpc[LPC+IDIM];
	real ZF_MEM    zirwfir[LPCW];
	real ZI_MEM    zirwiir[LPCW];

	real gain_input[LPCLG];

	real tap_mem[KPMAX+NPWSZ+IDIM]; /* Post-Filter Memory for syn. sp. */
	int pitch_period;
	real pitch_tap;
	real pitch_gain;

	real shpcoef[SPORDER+1], shzcoef[SPORDER+1], tiltz, fil_mem[PMSIZE];

/* declare array and its copy together with a semafor */

	real sf_coeff[LPC+1];
	real gp_coeff[LPCLG+1];
	real pwf_z_coeff[LPCW+1];
	real pwf_p_coeff[LPCW+1];
	real shape_energy[NCWD];
	real imp_resp[IDIM];
	
	real sf_coeff_next[LPC+1];
	real gp_coeff_next[LPCLG+1];
	real pwf_z_coeff_next[LPCW+1];
	real pwf_p_coeff_next[LPCW+1];
	real shape_energy_next[NCWD];
	real imp_resp_next[IDIM];
	
	int sf_coeff_obsolete_p;
	int gp_coeff_obsolete_p;
	int pwf_z_coeff_obsolete_p;
	int pwf_p_coeff_obsolete_p;
	int shape_energy_obsolete_p;
	int imp_resp_obsolete_p;
	
	real synspeech [QSIZE];		/* Synthesized Speech */
	real qspeech   [QSIZE];		/* Quantized   Speech */
	real log_gains[QSIZE/IDIM];	/* Logarithm of Gains */

	int VOLATILE ffase;
} sound_g728_t;

#define FFASE(N) if (this->ffase == N)

#define NEXT_FFASE this->ffase=(this->ffase==4?1:this->ffase+1)

#endif /*COMMON_H*/
