/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

// see below for adding new output modules
#include "../config.h"

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#define OUTPUT_INTERNAL
#include "output.h"
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) output_ ## x

int output_file_fmt = 0;
unsigned char **output_buf = NULL;
int output_fbs = 0;
/************ Prototypes ************/

void output_init();
void output_start();
void output_stop();
void output_end();
int output_write(int fbnr, int picnr);

/************************************/
#define PROTO(symbol) \
	extern int output_ ## symbol ## _fmts(); \
	extern const char *output_ ## symbol ## _fmt(int nr); \
	extern void output_ ## symbol ## _start(const char *format); \
	extern int output_ ## symbol ## _write(unsigned char *buffer, int picnr); \
	extern void output_ ## symbol ## _stop(); \
	extern void output_ ## symbol ## _configure(const char *format, const char *assignment); \
	extern void output_ ## symbol ## _inquire(const char *format, char *value);

#define ADD_MODULE(symbol) \
 { \
	output_ ## symbol ## _fmts, \
	output_ ## symbol ## _fmt, \
	output_ ## symbol ## _start, \
	output_ ## symbol ## _write, \
	output_ ## symbol ## _stop, \
	output_ ## symbol ## _configure, \
	output_ ## symbol ## _inquire, \
 }

struct OUTPUT_MODULE_STRUCT {
	int (*fmts)();
	const char *(*fmt)(int nr);
	void (*start)(const char *format);
	int (*write)(unsigned char *buffer, int picnr);
	void (*stop)();
	void (*configure)(const char *format, const char *assignment);
	void (*inquire)(const char *format, char *value);
};

struct OUTPUT_FORMAT_STRUCT {
	char name[80];
	int modulenum;
} *output_formats = NULL;

int OUTPUT_FORMATS = 0;

////////////////////////////////////////////////////////////////////////////////
///////////////////////// output module registration ///////////////////////////
////////////////////////////////////////////////////////////////////////////////

PROTO(pmm)
PROTO(jpg)
PROTO(gif)
PROTO(raw)
PROTO(pnm)
PROTO(H263)
PROTO(term)
PROTO(X)
PROTO(fb)
PROTO(Win32)

// Step 1: fill in structure with ADD_FORMAT
//         supply literal name, symbol name, color format (w/o OUTPUT_ prefix)
const struct OUTPUT_MODULE_STRUCT output_modules[] = {
	ADD_MODULE(pmm),
#ifdef HAVE_LIBJPEG
	ADD_MODULE(jpg),
#endif
#ifdef HAVE_LIBGIF
	ADD_MODULE(gif),
#endif
	ADD_MODULE(raw),
	ADD_MODULE(pnm),
	ADD_MODULE(H263),
#ifdef HAVE_LIBAA
	ADD_MODULE(term),
#endif
#ifdef HAVE_LIBX11
	ADD_MODULE(X),
#endif
#if defined(HAVE_LIBUSER32) && defined(HAVE_LIBGDI32)
	ADD_MODULE(Win32),
#endif
#ifdef HAVE_LINUX_FB_H
        ADD_MODULE(fb),
#endif
	{ NULL, NULL, NULL, NULL, NULL, NULL, NULL }
};

// Step 3: add the new object file to Makefile.in
// Step 4: create the source file, see output_jpg.c for an example

// finished
////////////////////////////////////////////////////////////////////////////////

START_OPTIONS
INTOPTION(file_fmt,0,OUTPUT_FORMATS-1);
OPTION(file_fmt nr) NOSET O_GET sprintf(value,"%i",OUTPUT_FORMATS);
OPTION(file_fmt str) O_SET {
	int i;
	for (i = 0; i < OUTPUT_FORMATS; i++) {
		if (!strcmp(output_formats[i].name,value)) {
			output_file_fmt = i;
			break;
		}
	}
	ASSERT(i < OUTPUT_FORMATS,(MSG("unknown output format '%s'"),value));
} O_GET {
	strcpy(value,output_formats[output_file_fmt].name);
}
INTOPTION(fbs,0,INT_MAX) if (set) {
	SAFE_FREE(output_buf);
	SAFE_CALLOC(output_buf,output_fbs,sizeof(void *));
}
PTROPTION(buf,output_fbs);
OPTION(data) O_SET (*output_modules[output_formats[output_file_fmt].modulenum].configure)(output_formats[output_file_fmt].name,value);
	     O_GET (*output_modules[output_formats[output_file_fmt].modulenum].inquire)(output_formats[output_file_fmt].name,value);
END_OPTIONS

int (*output_write_func)(unsigned char *buffer, int picnr);
void (*output_stop_func)() = NULL;

void output_init()
{
	int modulecount = 0;
	int fmtcount;
	OUTPUT_FORMATS = 0;
	
	while (output_modules[modulecount].start) {
		fmtcount = (*output_modules[modulecount].fmts)();
		OUTPUT_FORMATS += fmtcount;
		NOTNULL(output_formats = realloc(output_formats, OUTPUT_FORMATS*sizeof(struct OUTPUT_FORMAT_STRUCT)),
			(MSG("realloc failed for %i bytes"),OUTPUT_FORMATS*sizeof(struct OUTPUT_FORMAT_STRUCT)));
		while (fmtcount--) {
			output_formats[OUTPUT_FORMATS-fmtcount-1].modulenum = modulecount;
			strcpy(output_formats[OUTPUT_FORMATS-fmtcount-1].name,(*output_modules[modulecount].fmt)(fmtcount));
		}
		modulecount++;
	}

	transport_init();
}

void output_start()
{
	ASSERT(output_fbs,(INTMSG("image buffers not set")));

	output_write_func = output_modules[output_formats[output_file_fmt].modulenum].write;
	output_stop_func = output_modules[output_formats[output_file_fmt].modulenum].stop;

	(*output_modules[output_formats[output_file_fmt].modulenum].start)(output_formats[output_file_fmt].name);
}

void output_stop()
{
	if (output_stop_func) (*output_stop_func)();
	output_stop_func = NULL;
	transport_wstop();
}

void output_end()
{
	output_stop();
	SAFE_FREE(output_buf);
	SAFE_FREE(output_formats);
}

int output_write(int fbnr, int picnr)
{
	return (*output_write_func)(output_buf[fbnr],picnr);
}
