/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include "error.h"
#include "transport.h"
#include "color.h"
#include "configuration.h"
#define SYMBOL(x) output_raw_ ## x

/************ Prototypes ************/

void output_raw_start(const char *format);
int output_raw_write(char *buffer, int picnr);
void output_raw_stop();

/************************************/

int output_raw_fmts() { return 7; }
const char *output_raw_fmt(int nr) {
	switch (nr) {
	case 0: return "raw8";
	case 1: return "raw16";
	case 2: return "raw24";
	case 3: return "rawyuv";
	case 4: return "rawyuv422";
	case 5: return "rawgrey";
	default: return "yuv";
	}
}

int output_raw_width = 0, output_raw_height = 0, output_raw_bytes = 0;
START_DATA
INTOPTION(width,0,INT_MAX);
INTOPTION(height,0,INT_MAX);
OPTION(color_fmt) NOSET O_GET {
	if (!strcmp(format,"raw8")) strcpy(value,"PAL8");
	else if (!strcmp(format,"raw16")) strcpy(value,"RGB16");
	else if (!strcmp(format,"raw24")) strcpy(value,"RGB24S");
	else if (!strcmp(format,"rawyuv")) strcpy(value,"YUV420");
	else if (!strcmp(format,"rawyuv422")) strcpy(value,"YUV422S");
	else if (!strcmp(format,"rawgrey")) strcpy(value,"GREY8");
	else if (!strcmp(format,"yuv")) strcpy(value,"YUV420P");
	else {
		ERROR(INTMSG("unknown raw type: %s"),format);
		END(1);
	}
}
END_DATA

void output_raw_start(const char *format)
{
	char tmp[10];
	strcpy(tmp,"color_fmt");
	output_raw_inquire(format,tmp);
	output_raw_bytes = color_bytes(color_val(tmp),output_raw_width,output_raw_height);

	transport_wstart(format,output_raw_width,output_raw_height,1);
}

int output_raw_write(char *buffer, int picnr)
{
	transport_wframe(picnr);
	transport_write(buffer,output_raw_bytes);
	return 1;
}

void output_raw_stop()
{
}
