/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* This file based on: */
/************************************************************************
 *
 *  tmndec.c, main(), initialization, options for tmndecode (H.263 decoder)
 *  Copyright (C) 1995, 1996  Telenor R&D, Norway
 *
 *  Contacts:
 *  Robert Danielsen                  <Robert.Danielsen@nta.no>
 *
 *  Telenor Research and Development  http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 *  Copyright (C) 1997  University of BC, Canada
 *  Modified by: Michael Gallant <mikeg@ee.ubc.ca>
 *               Guy Cote <guyc@ee.ubc.ca>
 *               Berna Erol <bernae@ee.ubc.ca>
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 ************************************************************************/

/* Disclaimer of Warranty
 * 
 * These software programs are available to the user without any license fee
 * or royalty on an "as is" basis. The University of British Columbia
 * disclaims any and all warranties, whether express, implied, or
 * statuary, including any implied warranties or merchantability or of
 * fitness for a particular purpose.  In no event shall the
 * copyright-holder be liable for any incidental, punitive, or
 * consequential damages of any kind whatsoever arising from the use of
 * these programs.
 * 
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 * 
 * The University of British Columbia does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any
 * third-party patents.
 * 
 * Commercial implementations of H.263, including shareware, are subject to
 * royalty fees to patent holders.  Many of these patents are general
 * enough such that they are unavoidable regardless of implementation
 * design.
 * 
 */

/* modified to support annex O true B frames, mikeg@ee.ubc.ca
 * 
 * based on mpeg2decode, (C) 1994, MPEG Software Simulation Group and
 * mpeg2play, (C) 1994 Stefan Eckart <stefan@lis.e-technik.tu-muenchen.de>
 * 
 */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <limits.h>
#define READ_INTERNAL
#include "read.h"
#define ERROR_DEBUG
#include "error.h"
#include "tmndec/tmndec.h"
#define GLOBAL
#include "tmndec/global.h"
#define o(x) (getenv(#x)?atoi(getenv(#x)):x)
#include "configuration.h"
#define SYMBOL(x) read_H263_ ## x

/************ Prototypes ************/

void read_H263_start();
void read_H263_stop();
int read_H263_read(int fbnr, int picnr);
static inline void arrow(unsigned char *, int, int, int, int, unsigned char);
static inline void line(unsigned char *, int, int, int, int, unsigned char);

/************************************/

#define REFIDCT 0
#define EXPAND 0
#define POST_FILTER 0
#define CONCEALMENT 0
#define QUIET 0

int read_H263_end = 0;
int read_H263_gob = 0;
int read_H263_buffered_frame_temp_ref = 0;
int read_H263_first = 1;
int read_H263_pb_pending = 0;
int read_H263_update_buffered_frame = ON, read_H263_n = 0;

int read_H263MV_fd = -1;
char read_H263_vectors[PATH_MAX] = "";

#define MVBUF_MAX 32
short read_H263MV_buf[MVBUF_MAX][2];
int read_H263MV_bcur;

START_DATA
BOOLOPTION(expand);
STROPTION(vectors);
END_DATA

void read_H263_start()
{
	int i, size;
	transport_frstart(0);

	read_H263_end = 0;
	read_H263_verbose = 0;
	read_H263_refidct = 0;//o(REFIDCT);
	read_H263_quiet = 1; //o(QUIET);
	read_H263_post_filter = 1;//o(POST_FILTER);
	read_H263_concealment = 0;//o(CONCEALMENT);
	read_H263_ld = &read_H263_base;
	read_H263_initbits();
	read_H263_initoptions();
	read_H263_temp_ref = 0;
	read_H263_prev_non_disposable_temp_ref = read_H263_next_non_disposable_temp_ref = 0;
	do read_H263_gob = read_H263_getheader(); while (read_H263_gob != 1);
	read_width  = read_H263_coded_picture_width*(read_H263_expand+1);
	read_height = read_H263_coded_picture_height*(read_H263_expand+1);
	read_color_fmt = "YUV420P";

        read_H263_base_source_format = read_H263_source_format;

	/* clip table */
	SAFE_MALLOC(read_H263_clp,1024);
  
	read_H263_clp += 384;
	for (i = -384; i < 640; i++) read_H263_clp[i] = (i < 0) ? 0 : ((i > 255) ? 255 : i);

	/* MPEG-1 = TMN parameters */
	read_H263_matrix_coefficients = 5;

	read_H263_blk_cnt = 6;

	for (i=0; i<MAX_LAYERS; i++) read_H263_enhancement_layer_init[i] = OFF;

	for (i = 0; i < 3; i++) {
		if (i == 0) size = read_H263_coded_picture_width * read_H263_coded_picture_height;
		else size = read_H263_chrom_width * read_H263_chrom_height;
		/* Used for bidirectional and direct prediction mode for true B
		 * pictures, one for forward and one for backward. */
		SAFE_MALLOC(read_H263_next_I_P_frame[i],size);
		SAFE_MALLOC(read_H263_prev_I_P_frame[i],size);
		SAFE_MALLOC(read_H263_current_frame[i],size);
		SAFE_MALLOC(read_H263_bframe[i],size);
		SAFE_MALLOC(read_H263_tmp_f[i],size);
		SAFE_MALLOC(read_H263_tmp_b[i],size);
		SAFE_MALLOC(read_H263_buffered_frame[i],size);
		read_H263_prev_frame[i] = read_H263_current_frame[i];
	}

	for (i = 0; i < 3; i++) {
		if (i == 0) {
			size = (read_H263_coded_picture_width + 64) * (read_H263_coded_picture_height + 64);
			/* Stores framed version of previous inter-picture, luminance. */
			SAFE_MALLOC(read_H263_edgeframeorig[i],size);
			read_H263_edgeframe[i] = read_H263_edgeframeorig[i] + (read_H263_coded_picture_width + 64) * 32 + 32;

			/* Required for true B frames backward, bidir, and direct
			 * prediction. Stores future inter-picture, luminance */
			SAFE_MALLOC(read_H263_nextedgeframeorig[i],size);
			read_H263_nextedgeframe[i] = read_H263_nextedgeframeorig[i] + (read_H263_coded_picture_width + 64) * 32 + 32;
		} else {
			size = (read_H263_chrom_width + 32) * (read_H263_chrom_height + 32);

			/* Stores framed version of previous inter-picture, chrominance. */
			SAFE_MALLOC(read_H263_edgeframeorig[i],size);
			read_H263_edgeframe[i] = read_H263_edgeframeorig[i] + (read_H263_chrom_width + 32) * 16 + 16;
			/* Required for true B frames backward, bidir, and direct
			 * prediction. Stores future inter-picture, chrominance. */
			SAFE_MALLOC(read_H263_nextedgeframeorig[i],size);
			read_H263_nextedgeframe[i] = read_H263_nextedgeframeorig[i] + (read_H263_chrom_width + 32) * 16 + 16;
		}
	}

	/* IDCT */
	if (read_H263_refidct) read_H263_init_idctref();
	else read_H263_init_idct();

	read_H263_prev_mv_outside_frame = read_H263_prev_sac = read_H263_prev_adv_pred = read_H263_prev_aic = read_H263_prev_df = 0;
	read_H263_prev_slice_struct = read_H263_prev_rps = read_H263_prev_isd = read_H263_prev_aivlc = read_H263_prev_mq = 0;
	read_H263_prev_4mv = read_H263_prev_long_vectors = read_H263_prev_obmc = 0;

	read_H263_gotframe = 0;
	i = 0;
	
	read_H263_stop_decoder = 0;
	read_H263_first = 1;
	read_H263_pb_pending = 0;
	read_H263_update_buffered_frame = ON;
	read_H263_n = 0;

}

void read_H263_stop()
{
	read_H263_clp -= 384;
	SAFE_FREE(read_H263_clp);
}

int read_H263_picnr;
int read_H263_read(int fbnr, int picnr)
{
	unsigned char *frame_to_display[3], *exnewframe[3];

	ASSERT(read_H263_end < 2,(MSG("end of input stream")));

	exnewframe[0] = read_ptrs[fbnr][0];
	exnewframe[1] = exnewframe[0] + read_width*read_height;
	exnewframe[2] = exnewframe[1] + read_width*read_height/4;
	if (read_H263_pb_pending) {
		if (read_H263_expand) {
			read_H263_interpolate_image(read_H263_bframe[0], exnewframe[0], read_H263_coded_picture_width, read_H263_coded_picture_height);
			read_H263_interpolate_image(read_H263_bframe[1], exnewframe[1], read_H263_chrom_width, read_H263_chrom_height);
			read_H263_interpolate_image(read_H263_bframe[2], exnewframe[2], read_H263_chrom_width, read_H263_chrom_height);
		} else {
			memcpy(exnewframe[0],read_H263_bframe[0],read_H263_coded_picture_width*read_H263_coded_picture_height);
			memcpy(exnewframe[1],read_H263_bframe[1],read_H263_chrom_width*read_H263_chrom_height);
			memcpy(exnewframe[2],read_H263_bframe[2],read_H263_chrom_width*read_H263_chrom_height);
		}
		read_H263_pb_pending = 0;
	} else {
		if (!read_H263_end) {
			transport_frframe(&read_H263_n);

			read_H263_gotframe = 0;
			do {
				if (read_H263_stop_decoder) {
					while (read_H263_pict_type != PCT_INTRA) read_H263_gob = read_H263_getheader();
					read_H263_stop_decoder = 0;
				} else {
					if (!read_H263_first) read_H263_gob = read_H263_getheader();
				}
				if (read_H263_gob) read_H263_getpicture(&read_H263_n, read_H263_gob);
			} while(!read_H263_gotframe && read_H263_gob);
		} else {
			read_H263_end++;
		}

	if (read_H263_first) {
		read_H263_first = 0;
//		read_H263_read(fbnr,picnr);
	}

//		if (read_H263_end || ((read_H263_temp_ref < read_H263_buffered_frame_temp_ref) && (PCT_B == read_H263_pict_type))) {
			/* Do not store current frame. */
			read_H263_update_buffered_frame = OFF;
			frame_to_display[0] = read_H263_current_frame[0];
			frame_to_display[1] = read_H263_current_frame[1];
			frame_to_display[2] = read_H263_current_frame[2];
			picnr = read_H263_picnr; //temp_ref;
//		} else {
//			frame_to_display[0] = read_H263_buffered_frame[0];
//			frame_to_display[1] = read_H263_buffered_frame[1];
//			frame_to_display[2] = read_H263_buffered_frame[2];
//			/* Store current frame. */
//			read_H263_update_buffered_frame = ON;
//			picnr = read_H263_buffered_frame_temp_ref;
//		}

		/* Post Filter */
		if (read_H263_post_filter) read_H263_PostFilter(frame_to_display[0], frame_to_display[1], frame_to_display[2], read_H263_coded_picture_width, read_H263_coded_picture_height);
		if (read_H263_expand) {
			/* display this image and keep frame already in storage where it is. */
			read_H263_interpolate_image(frame_to_display[0], exnewframe[0], read_H263_coded_picture_width, read_H263_coded_picture_height);
			read_H263_interpolate_image(frame_to_display[1], exnewframe[1], read_H263_chrom_width, read_H263_chrom_height);
			read_H263_interpolate_image(frame_to_display[2], exnewframe[2], read_H263_chrom_width, read_H263_chrom_height);
		
		} else {
			memcpy(exnewframe[0],frame_to_display[0],read_H263_coded_picture_width*read_H263_coded_picture_height);
			memcpy(exnewframe[1],frame_to_display[1],read_H263_chrom_width*read_H263_chrom_height);
			memcpy(exnewframe[2],frame_to_display[2],read_H263_chrom_width*read_H263_chrom_height);
		}
		if (read_H263_pb_frame) {
			read_H263_pb_pending = 1;
			return picnr;
		}
	}
	if (read_H263_update_buffered_frame) {
		read_H263_buffered_frame_temp_ref = read_H263_temp_ref + (((read_H263_temp_ref-read_H263_buffered_frame_temp_ref) < 0) ? 256 : 0);
		memcpy(read_H263_buffered_frame[0], read_H263_current_frame[0], read_H263_coded_picture_width * read_H263_coded_picture_height);
		memcpy(read_H263_buffered_frame[1], read_H263_current_frame[1], read_H263_chrom_width * read_H263_chrom_height);
		memcpy(read_H263_buffered_frame[2], read_H263_current_frame[2], read_H263_chrom_width * read_H263_chrom_height);
	}

#ifdef VISI
	if (*read_H263_vectors) {
		int i, lx = 0, ly = 0;
		if ((read_H263MV_fd = open(read_H263_vectors,O_RDONLY|O_BINARY)) >= 0) {
			read(read_H263MV_fd,read_H263MV_buf,sizeof(read_H263MV_buf));
			close(read_H263MV_fd);
			for (i = 1; i < MVBUF_MAX; lx = read_H263MV_buf[i][0], ly = read_H263MV_buf[i][1], i++) {
				if (read_H263MV_buf[i][0] == 0 && read_H263MV_buf[i][1] == 0) continue;
				read_H263MV_buf[i][0] = read_H263MV_buf[i][0]*read_width/read_H263MV_buf[0][0];
				read_H263MV_buf[i][1] = read_H263MV_buf[i][1]*read_width/read_H263MV_buf[0][1];
				if (lx || ly) arrow(exnewframe[0],lx,ly,read_H263MV_buf[i][0],read_H263MV_buf[i][1],255-3*MVBUF_MAX+3*i);
			}
		}
	}
#endif
	return picnr;
}

static inline void arrow(unsigned char *buf, int xs, int ys, int xd, int yd, unsigned char c)
{
	int dx = xd-xs;
	int dy = yd-ys;
	double alpha;
	line(buf,xs,ys,xd,yd,c);
	if (dx != 0) alpha = atan((float)dy/(float)dx);
	else alpha = dy > 0? M_PI/2.0 : -M_PI/2.0;
	if (dx < 0) alpha+=M_PI;
	alpha += M_PI*0.75+0.7;
	line(buf,xd+(cos(alpha)-sin(alpha))*4,yd+(sin(alpha)+cos(alpha))*4,xd,yd,c);
	alpha -= 1.4;
	line(buf,xd+(cos(alpha)-sin(alpha))*4,yd+(sin(alpha)+cos(alpha))*4,xd,yd,c);
}

static inline void line(unsigned char *buf, int xs, int ys, int xd, int yd, unsigned char c)
{
	int xc, yc, xi, yi;
//	if (read_H263_expand) {
//		xs *= 2;
//		ys *= 2;
//		xd *= 2;
//		yd *= 2;
//	}
	xc = xs;
	yc = ys;
	xi = xs < xd ?  1 : -1;
	yi = ys < yd ?  1 : -1;
	*(buf+yc*read_width+xc) = c;
	while (xc*xi < xd*xi || yc*yi < yd*yi) {
		if ((xc == xd) || ((float)(xi*yi*(yd-yc))/(float)((xd-xc)) > (float)(xi*yi*(yd-ys))/(float)((xd-xs)))) yc += yi;
		else xc += xi;
		*(buf+yc*read_width+xc) = c;
	}
		
}
