/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

// see below for adding new output modules
#include "../config.h"

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "transport.h"
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) read_sound_ ## x

unsigned char **read_sound_buf = NULL;
int read_sound_fbs = 512;
int read_sound_file_fmt = 0;
/************ Prototypes ************/

void read_sound_init();
void read_sound_start();
void read_sound_stop();
void read_sound_end();
int read_sound_get(int *fbnr, int *picnr);
void read_sound_setnext(int fbnr, int picnr);

/************************************/
#define PROTO(symbol) \
	extern int read_sound_ ## symbol ## _get(int *fbnr, int *picnr); \
	extern void read_sound_ ## symbol ## _setnext(int fbnr, int picnr); \
	extern void read_sound_ ## symbol ## _start(const char *format); \
	extern void read_sound_ ## symbol ## _stop(); \
	extern void read_sound_ ## symbol ## _configure(const char *format, const char *assignment); \
	extern void read_sound_ ## symbol ## _inquire(const char *format, char *value);

#define ADD_MODULE(symbol) { \
	read_sound_ ## symbol ## _get, \
	read_sound_ ## symbol ## _setnext, \
	read_sound_ ## symbol ## _start, \
	read_sound_ ## symbol ## _stop, \
	read_sound_ ## symbol ## _configure, \
	read_sound_ ## symbol ## _inquire, \
	#symbol }

struct READ_SOUND_MODULE_STRUCT {
	int (*get)(int *fbnr, int *picnr);
	void (*setnext)(int fbnr, int picnr);
	void (*start)(const char *format);
	void (*stop)();
	void (*configure)(const char *format, const char *assignment);
	void (*inquire)(const char *format, char *value);
	const char *name;
};

int READ_SOUND_FORMATS = 0;

////////////////////////////////////////////////////////////////////////////////
///////////////////////// output module registration ///////////////////////////
////////////////////////////////////////////////////////////////////////////////

// Step 1: add prototypes
PROTO(dsp)
PROTO(wav)
PROTO(g728)

// Step 2: fill in structure with ADD_FORMAT (supply symbol prefix)
const struct READ_SOUND_MODULE_STRUCT read_sound_formats[] = {
	ADD_MODULE(dsp),
	ADD_MODULE(wav),
	ADD_MODULE(g728),
	{ NULL, NULL, NULL, NULL, NULL, NULL, NULL }
};

// Step 3: add the new object file to Makefile.in
// Step 4: create the source file, see read_sound_wav.c for an example

// finished
////////////////////////////////////////////////////////////////////////////////

START_OPTIONS
INTOPTION(file_fmt,0,READ_SOUND_FORMATS-1);
OPTION(file_fmt nr) NOSET O_GET sprintf(value,"%i",READ_SOUND_FORMATS);
OPTION(file_fmt str) O_SET {
	int i;
	for (i = 0; i < READ_SOUND_FORMATS; i++) {
		if (!strcmp(read_sound_formats[i].name,value)) {
			read_sound_file_fmt = i;
			break;
		}
	}
	ASSERT(i < READ_SOUND_FORMATS,(MSG("unknown sound input format '%s'"),value));
} O_GET {
	strcpy(value,read_sound_formats[read_sound_file_fmt].name);
}
INTOPTION(fbs,4,INT_MAX);
PTRINFO(buf,read_sound_fbs);
OPTION(data) O_SET (*read_sound_formats[read_sound_file_fmt].configure)(read_sound_formats[read_sound_file_fmt].name,value);
             O_GET (*read_sound_formats[read_sound_file_fmt].inquire)(read_sound_formats[read_sound_file_fmt].name,value);
END_OPTIONS

 int (*read_sound_get_func)(int *fbnr, int *picnr);
 void (*read_sound_setnext_func)(int fbnr, int picnr);
 void (*read_sound_stop_func)() = NULL;

void read_sound_init()
{
	int modulecount = 0;
//	int fmtcount;
	READ_SOUND_FORMATS = 0;

	while (read_sound_formats[modulecount].start) {
//		fmtcount = (*read_sound_modules[modulecount].fmts)();
//		READ_SOUND_FORMATS += fmtcount;
//		NOTNULL(read_sound_formats = realloc(read_sound_formats, READ_SOUND_FORMATS*sizeof(struct READ_SOUND_FORMAT_STRUCT)),
//			(MSG("realloc failed for %i bytes"),READ_SOUND_FORMATS*sizeof(struct READ_SOUND_FORMAT_STRUCT)));
//		while (fmtcount--) {
//			read_sound_formats[READ_SOUND_FORMATS-fmtcount-1].modulenum = modulecount;
//			strcpy(read_sound_formats[READ_SOUND_FORMATS-fmtcount-1].name,(*read_sound_modules[modulecount].fmt)(fmtcount));
//		}
		modulecount++;
	}

	READ_SOUND_FORMATS = modulecount;
	transport_init();
}

void read_sound_start()
{
	char tmp[10];
	int i, bufsize;

	read_sound_inquire("file_fmt str",tmp);
	transport_srdetect(tmp);
	read_sound_configure("file_fmt str",tmp);

	(*read_sound_formats[read_sound_file_fmt].start)(read_sound_formats[read_sound_file_fmt].name);
	strcpy(tmp,"bufsize");
	(*read_sound_formats[read_sound_file_fmt].inquire)(read_sound_formats[read_sound_file_fmt].name,tmp);
	bufsize = atoi(tmp);

	SAFE_CALLOC(read_sound_buf,01,read_sound_fbs*sizeof(char *));
	SAFE_MALLOC(read_sound_buf[0],bufsize*read_sound_fbs);
	for (i = 1; i < read_sound_fbs; i++) read_sound_buf[i] = read_sound_buf[i-1]+bufsize;

	read_sound_get_func = read_sound_formats[read_sound_file_fmt].get;
	read_sound_setnext_func = read_sound_formats[read_sound_file_fmt].setnext;
	read_sound_stop_func = read_sound_formats[read_sound_file_fmt].stop;

}

void read_sound_stop()
{
	if (read_sound_stop_func) (*read_sound_stop_func)();
	read_sound_stop_func = NULL;
}

void read_sound_end()
{
	read_sound_stop();
	SAFE_FREE(read_sound_buf);
}

int read_sound_get(int *fbnr, int *picnr)
{
	return (*read_sound_get_func)(fbnr,picnr);
}

void read_sound_setnext(int fbnr, int picnr)
{
	(*read_sound_setnext_func)(fbnr, picnr);
}
