/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*************************************************************************/
/*                                                                       */
/*                            LD-CELP  G.728                             */
/*                                                                       */
/*    Low-Delay Code Excitation Linear Prediction speech compression.    */
/*                                                                       */
/*    Code edited by Michael Concannon.                                  */
/*    Based on code written by Alex Zatsman, Analog Devices 1993         */
/*                                                                       */
/*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "ld-celp/parm.h"
#include "ld-celp/prototyp.h"
#include "ld-celp/common.h"
#include "error.h"
#include "configuration.h"
#include "read_sound.h"
#include "transport.h"
#define SYMBOL(x) read_sound_g728_ ## x

/************ Prototypes ************/

void read_sound_g728_setnext(int fbnr, int picnr);
int read_sound_g728_get(int *fbnr, int *picnr);
void read_sound_g728_start();
void read_sound_g728_stop();
static void decode_vector(sound_g728_t *this);
static void adapt_decoder(sound_g728_t *this);
static inline int do_sync();
static inline unsigned char do_read();

/************************************/

int read_sound_g728_postfiltering_p = 1;
int read_sound_g728_rate     = 8000;
int read_sound_g728_bits     = 16;
int read_sound_g728_channels = 1;
int read_sound_g728_bufsize = 0;
static int synced = 0;

START_DATA
INTOPTION(rate,4000,48000);
INTOPTION(bits,8,16);
INTOPTION(channels,1,4);
INTINFO(bufsize);
INTOPTION(postfiltering_p,0,1);
END_DATA

static sound_g728_t *data = NULL;

void read_sound_g728_setnext(int fbnr, int picnr)
{
}

int read_sound_g728_get(int *fbnr, int *picnr)
{
	int i, j;
	static int nextpic = 0;
	real *ibuf, tmp;
	signed short *obuf;

	*fbnr = 0;
	transport_sfrframe(picnr);
	transport_srlseek(((read_sound_g728_bufsize+7)/8)*(*picnr-nextpic),SEEK_CUR);
	*picnr = do_sync();
	nextpic = *picnr+1;

	obuf = (signed short *)read_sound_buf[0];
	for (i = 0; i < read_sound_g728_rate/125; i++) {
		if (data->w_vec_start >= QSIZE) data->w_vec_start = 0;
		data->w_vec_end = data->w_vec_start;
		data->vector_end = data->thequeue + data->w_vec_end;
		decode_vector(data);
		data->w_vec_end = data->w_vec_start + IDIM;
		if (data->w_vec_end >= QSIZE) data->w_vec_end = 0;
		ibuf = data->thequeue + data->w_vec_end;
		for (j = 0; j < IDIM; j++) {
			tmp = ibuf[j]/0.125;
			if (tmp > 0.0) {
				if (tmp > 32767.0) tmp = 32767.0;
				else tmp += 0.5;
			} else {
				if (tmp < -32768.0) tmp = -32768.0;
				else tmp -= 0.5;
			}
			*obuf = (signed short)tmp;
			obuf++;
		}
		adapt_decoder(data);
		data->w_vec_start += IDIM;
	}
	data->buf = 0;
	data->bits = 0;
	return read_sound_g728_bufsize;
}

void read_sound_g728_start()
{
	read_sound_g728_bits = 16;
	read_sound_g728_channels = 1;
	SAFE_CALLOC(data, 01, sizeof(*data));

	transport_sfrstart("g728");

	read_sound_g728_bufsize = read_sound_g728_rate/25*2;
	data->ffase = 1;
	data->pitch_period = 50;
	data->pitch_tap=0.0;
	data->pitch_gain=1;
	sound_g728_init_bsf_adapter(data,data->sf_coeff);
	data->sf_coeff_next[0] = 1.0;
	data->sf_coeff_obsolete_p = 0;
	sound_g728_init_gain_adapter(data,data->gp_coeff);
	sound_g728_init_gain_buf(data);
	data->gp_coeff_next[0] = 1.0;
	data->gp_coeff_next[1] = -1.0;
	data->gp_coeff_obsolete_p = 0;
	data->vector_end=data->thequeue;
}

void read_sound_g728_stop()
{
	SAFE_FREE(data);
}

static void decode_vector(sound_g728_t *this)
{
	int ix, tmp;	/* Computed Codebook Index */
	int lgx;	/* Log Gains INdex */
	unsigned char c;
	int b;
	static real zero_response[IDIM], cb_vec[IDIM];
	static real pf_speech[IDIM];
	real qs[NUPDATE*IDIM];
	
	static real	gain = 1.0;

	this->w_vec_end = this->vector_end - this->thequeue;
	this->d_vec_start = this->w_vec_end + IDIM;
	if (this->d_vec_start >= QSIZE) this->d_vec_start -= QSIZE;
	
	b = this->bits;
	ix = this->buf;
	while (b < 10) {
		c = do_read();
		if (synced) return;
		tmp = c;
		tmp <<= b;
		ix |= tmp;
		b += 8;
	}
	b -= 10;
	this->buf = ix >> 10;
	this->bits = b;
	ix &= 1023;

	UPDATE(this->sf_coeff);
	sound_g728_zresp(this, zero_response);
	sound_g728_cb_excitation(this, ix, cb_vec);
	UPDATE(this->gp_coeff);
	gain = sound_g728_predict_gain(this);
	sig_scale(gain, cb_vec, this->qspeech+this->d_vec_start);
	lgx = this->d_vec_start/IDIM;
	sound_g728_update_gain(this, this->qspeech+this->d_vec_start, this->log_gains+lgx);
	sound_g728_mem_update(this, this->qspeech+this->d_vec_start, this->synspeech+this->d_vec_start);
	this->d_vec_end = this->d_vec_start + IDIM;
	if (this->d_vec_end >= QSIZE) this->d_vec_end -= QSIZE;
	if (read_sound_g728_postfiltering_p) {
		sound_g728_inv_filter(this, this->synspeech+this->d_vec_start);
		FFASE(3) {
			CIRCOPY(qs, this->synspeech, this->d_vec_end, NUPDATE*IDIM, QSIZE);
			sound_g728_psf_adapter(this, qs);
		}
		FFASE(1)
			sound_g728_compute_sh_coeff(this);
		sound_g728_postfilter(this, this->synspeech+this->d_vec_start, pf_speech);
		RCOPY(pf_speech, this->thequeue+this->d_vec_start, IDIM);
	} else {
		RCOPY(this->synspeech+this->d_vec_start, this->thequeue+this->d_vec_start, IDIM);
	}
	NEXT_FFASE;
}

static void adapt_decoder(sound_g728_t *this)
{
	real synth [NUPDATE*IDIM],
		lg    [NUPDATE];
	int gx; /* gain index */
	
    
	FFASE(1) {     
		CIRCOPY(synth, this->synspeech, this->d_vec_end, NUPDATE*IDIM, QSIZE);
		sound_g728_bsf_adapter(this, synth, this->sf_coeff_next);
	}
	FFASE(2) {
		gx = this->d_vec_end / IDIM;
		CIRCOPY(lg, this->log_gains, gx, NUPDATE, QSIZE/IDIM);
		sound_g728_gain_adapter(this, lg, this->gp_coeff_next);
		this->gp_coeff_obsolete_p = 1;
	}
	FFASE(3) this->sf_coeff_obsolete_p = 1;
}

static inline int do_sync()
{
	unsigned char c = 1;
	int i;

	if (!synced) do {
		while (c) transport_sread(&c,1);
		for (i = 1; i < 4; i++) {
			transport_sread(&c,1);
			if (c) break;
		}
	} while (i < 4);
	return transport_sread_int();
}

static inline unsigned char do_read()
{
	unsigned char c;
	int i;

	if (synced) return 0;
	transport_sread(&c,1);
	if (!c) {
		synced = 1;
		for (i = 1; i < 4; i++) {
			transport_sread(&c,1);
			if (c) synced = 0;
		}
		c = 0;
	}

	return c;
}
