/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "../config.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#endif
#ifdef HAVE_LIBWINMM
#include <windows.h>
#include <mmsystem.h>
#undef ERROR
#endif
#include <errno.h>
#include "read_sound.h"
#define ERROR_DEBUG
#include "error.h"
#include "configuration.h"
#define SYMBOL(x) sound_dsp_ ## x

#ifdef __CYGWIN__
// these are all kludges, because cygwin doesn't yet have complete multithreading
#define PTH_T(x) &x
#define pthread_exit(x) my_pthread_exit(x)
#endif

/************ Prototypes ************/

#ifdef HAVE_LIBWINMM
static inline void free_one(MSG msg);
static inline int play_one();
DWORD WINAPI output_dsp_messageloop(LPVOID sem);
#endif
void sound_dsp_start();
void sound_dsp_write(unsigned char *buffer, int picnr, int len);
void sound_dsp_stop();

/************************************/

int sound_dsp_rate     = 44100;
int sound_dsp_bits     = 16;
int sound_dsp_channels = 2;

#ifdef HAVE_SYS_SOUNDCARD_H
char sound_dsp_device[PATH_MAX] = "/dev/dsp";
int sound_dsp_fdsp = -1;
#endif

#ifdef HAVE_LIBWINMM
extern void output_Win32_display(int picnr);
extern void output_Win32_setsync(int i);

#define BUFNUM 16
#define THRESHOLD 16
HWAVEOUT sound_dsp_handle = 0;
WAVEHDR sound_dsp_bufhdr[BUFNUM];
int sound_dsp_picnr[BUFNUM];
HANDLE sound_dsp_hth = 0;
int sound_dsp_curbuf = INT_MAX;
int sound_dsp_loaded = 0;
#endif

START_DATA
#ifdef HAVE_SYS_SOUNDCARD_H
STROPTION(device);
#endif
INTOPTION(rate,4000,48000);
INTOPTION(bits,8,16);
INTOPTION(channels,1,4);
END_DATA

#if !defined(HAVE_SYS_SOUNDCARD_H) && defined(HAVE_LIBWINMM)
static inline void free_one(MSG msg)
{
	int rc;
	ASSERT((rc=waveOutUnprepareHeader((void *)msg.wParam,(void *)msg.lParam,sizeof(sound_dsp_bufhdr[0]))) == MMSYSERR_NOERROR, (MSG("waveOutPrepareHeader failed (%i)"),rc));
	rc = ((int)msg.lParam-(int)sound_dsp_bufhdr)/sizeof(sound_dsp_bufhdr[0]);
	output_Win32_display(sound_dsp_picnr[rc]);
	sound_dsp_picnr[rc] = -1;
}

static inline int play_one()
{
	int rc, last, i;

	last = 0;
	for (i = 0; i < BUFNUM; i++) {
		if (sound_dsp_picnr[i]-10 > sound_dsp_curbuf) sound_dsp_picnr[i] = -1;
		if (sound_dsp_picnr[i] >= 0 &&
		    (sound_dsp_picnr[i] < sound_dsp_picnr[last] ||
			sound_dsp_picnr[last] < 0)) last = i;
	}
	if (sound_dsp_picnr[last] >= 0) {
		i = sound_dsp_picnr[last];
		if (sound_dsp_handle) {
			sound_dsp_bufhdr[last].dwFlags = 0;
			ASSERT((rc=waveOutPrepareHeader(sound_dsp_handle,&(sound_dsp_bufhdr[last]),sizeof(sound_dsp_bufhdr[0]))) == MMSYSERR_NOERROR, (MSG("waveOutPrepareHeader %i failed (%i)"),last,rc));
			ASSERT((rc=waveOutWrite(sound_dsp_handle,&(sound_dsp_bufhdr[last]),sizeof(sound_dsp_bufhdr[0]))) == MMSYSERR_NOERROR, (MSG("waveOutWrite failed (%i)"),rc));
			sound_dsp_picnr[last] = -i-2;
		} else sound_dsp_picnr[last] = -1;
		output_Win32_display(i);
		return 1;
	}
	return 0;
}

static int done = 0, written = 0;

DWORD WINAPI output_dsp_messageloop(LPVOID sem)
{
	int rc, last = 0, i, n;
	MSG msg;
	done = 0;
	written = 0;
	msg.message = 0;
        do {
		if (msg.message != MM_WOM_CLOSE /*&& msg.message != MM_WOM_OPEN*/) {
			if (msg.message == MM_WOM_DONE) {
				free_one(msg);
				done++;
			}
			if (sound_dsp_loaded-written > 0) {
				n = play_one();
				if (!sound_dsp_handle) done += n;
				written += n;
			} else {
				if (sound_dsp_loaded == -1) continue;
				if (sound_dsp_loaded-written < THRESHOLD) {
					if (done < written) continue;
					while (sound_dsp_loaded-written < THRESHOLD && sound_dsp_loaded != -1) Sleep(10);
				}
				for (i = 0; i < THRESHOLD; i++) {
					n = play_one();
					if (!sound_dsp_handle) done += n;
					written += n;
				}
			}
		}
	} while (GetMessage(&msg,NULL,0,0));
	return 0;
}

static DWORD tid;
static WAVEFORMATEX fmt;

void sound_dsp_restart()
{
	int rc;

	if (sound_dsp_handle) return;

	fmt.wBitsPerSample = sound_dsp_bits;
	fmt.wFormatTag = WAVE_FORMAT_PCM;
	fmt.nChannels = sound_dsp_channels;
	fmt.nSamplesPerSec = sound_dsp_rate;
	fmt.nBlockAlign =  sound_dsp_channels * sound_dsp_bits/8;
	fmt.nAvgBytesPerSec = fmt.nBlockAlign * sound_dsp_rate;
	fmt.cbSize = 0;

	if (waveOutOpen(&sound_dsp_handle,WAVE_MAPPER,(WAVEFORMATEX *)&fmt,tid,0,CALLBACK_THREAD) == MMSYSERR_NOERROR)
		output_Win32_setsync(1);
	else
		output_Win32_setsync(-1);
}

void sound_dsp_restop()
{
	if (sound_dsp_handle) {
		HWAVEOUT h = sound_dsp_handle;
		output_Win32_setsync(-1);

		sound_dsp_handle = 0;
		waveOutReset(h);
		while (written > done) Sleep(10);
		waveOutClose(h);
	}
}
#endif

void sound_dsp_start()
{
#ifdef HAVE_SYS_SOUNDCARD_H
	int tmp;
	SAFE_OPEN(sound_dsp_fdsp,sound_dsp_device,O_WRONLY);
	tmp = (sound_dsp_rate/25*sound_dsp_bits/8*sound_dsp_channels)/256+2;
	tmp = (tmp << 16) | 0x0008;
	NOTNEG1(ioctl(sound_dsp_fdsp, SNDCTL_DSP_SETFRAGMENT, &tmp),(ERRMSG("fragment %x"),tmp));
	tmp = AFMT_S16_LE;
	if (sound_dsp_bits == 8) tmp = AFMT_U8;
	NOTNEG1(ioctl(sound_dsp_fdsp, SNDCTL_DSP_SETFMT, &tmp),(ERRMSG("sound format %i"),tmp));
	NOTNEG1(ioctl(sound_dsp_fdsp, SNDCTL_DSP_CHANNELS, &sound_dsp_channels),(ERRMSG("%i sound channels"),sound_dsp_channels));
	NOTNEG1(ioctl(sound_dsp_fdsp, SNDCTL_DSP_SPEED, &sound_dsp_rate),(ERRMSG("sampling rate %i"),sound_dsp_rate));
#else
#ifdef HAVE_LIBWINMM
	int rc, i;

	sound_dsp_loaded = 0;
        NOTNULL(sound_dsp_hth = CreateThread(NULL,0,output_dsp_messageloop,(void *)NULL,0,&tid),
		(MSG("could not create message thread")));
	sound_dsp_restart();

	for (i = 0; i < BUFNUM; i++) {
		SAFE_MALLOC(sound_dsp_bufhdr[i].lpData,fmt.nAvgBytesPerSec/25);
		sound_dsp_bufhdr[i].dwBufferLength = fmt.nAvgBytesPerSec/25;
		sound_dsp_bufhdr[i].dwFlags = 0;
		sound_dsp_bufhdr[i].dwFlags = WHDR_DONE;
		sound_dsp_picnr[i] = -1;
	}

	sound_dsp_curbuf = INT_MAX;

#else
	ASSERT(0,(MSG("no digital sound device configured")));
#endif
#endif
}

void sound_dsp_write(unsigned char *buffer, int picnr, int len)
{
#ifdef HAVE_SYS_SOUNDCARD_H
	SAFE_WRITE(sound_dsp_fdsp,buffer,len,"sound data");
#else
#ifdef HAVE_LIBWINMM
	int new, rc;
	if (!sound_dsp_handle) return;
	do {
		for (new = 0; new < BUFNUM; new++) {
			if (sound_dsp_picnr[new] == -1) break;
		}
		if (new < BUFNUM) break;
		waveOutReset(sound_dsp_handle);
		return;
	} while (0==0);
	sound_dsp_curbuf = picnr;
	memcpy(sound_dsp_bufhdr[new].lpData,buffer,len);
	sound_dsp_bufhdr[new].dwFlags &= ~WHDR_DONE;
	sound_dsp_picnr[new] = picnr;
	sound_dsp_loaded++;
#else
	ASSERT(0,(MSG("no digital sound device configured")));
#endif
#endif
}

void sound_dsp_stop()
{
#ifdef HAVE_SYS_SOUNDCARD_H
	if (sound_dsp_fdsp > 0) {
		SAFE_CLOSE(sound_dsp_fdsp,sound_dsp_device);
		sound_dsp_fdsp = -1;
	}
#else
#ifdef HAVE_LIBWINMM
	int i;
	sound_dsp_restop();
	for (i = 0; i < BUFNUM; i++) {
		SAFE_FREE(sound_dsp_bufhdr[i].lpData);
	}
#else

#endif
#endif
}

