/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*************************************************************************/
/*                                                                       */
/*                            LD-CELP  G.728                             */
/*                                                                       */
/*    Low-Delay Code Excitation Linear Prediction speech compression.    */
/*                                                                       */
/*    Code edited by Michael Concannon.                                  */
/*    Based on code written by Alex Zatsman, Analog Devices 1993         */
/*                                                                       */
/*************************************************************************/


/* Real multitasking works on a DSP target only, on other platform
   it is simulated */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "ld-celp/parm.h"
#include "ld-celp/prototyp.h"
#include "ld-celp/common.h"
#include "error.h"
#include "configuration.h"
#include "sound.h"
#include "transport.h"
#define SYMBOL(x) sound_g728_ ## x

/************ Prototypes ************/

void sound_g728_write(signed short *buffer, int picnr, int len);
void sound_g728_start();
void sound_g728_stop();
static void encode_vector(sound_g728_t *this);
static void adapt_frame(sound_g728_t *this);
static inline void do_write(unsigned char c);

/************************************/


int sound_g728_rate     = 44100;
int sound_g728_bits     = 16;
int sound_g728_channels = 1;

START_DATA
INTOPTION(rate,4000,48000);
INTOPTION(bits,8,16);
INTOPTION(channels,1,4);
END_DATA

static sound_g728_t *data = NULL;

void sound_g728_write(signed short *buffer, int picnr, int len)
{
	static int rest = 0;
	int i, j;

	transport_swframe(picnr);
	transport_swrite_int(0);
	transport_swrite_int(picnr);
	for(i = 0; (i+IDIM) <= (sound_g728_rate/25); i += IDIM) {
		for (j = 0; j < IDIM; j++) {
			if (rest) rest--;
			else if ((i+j)*2 < len) {
				data->thequeue[(i+j)%QSIZE] = ((real)*(buffer++))*0.125;
			} else {
				data->thequeue[(i+j)%QSIZE] = 0.0;
			}
		}
		data->vector_end = data->thequeue+i%QSIZE+IDIM;
		encode_vector(data);
		adapt_frame(data);
	}

	while (i*2 < len) {
		data->thequeue[rest] = ((real)*(buffer++))*0.125;
		rest++;
		i++;
	}
	if (data->bits) {
		do_write(data->buf);
		data->bits = 0;
		data->buf = 0;
	}
}

void sound_g728_start()
{
	int i;

	ASSERT(sound_g728_bits == 16,(MSG("G.728 needs 16 bit input")));
	ASSERT(sound_g728_channels == 1,(MSG("G.728 only supports mono sound")));
	SAFE_CALLOC(data, 01, sizeof(*data));

	data->ffase = 1;
	data->pitch_period = 50;
	data->pitch_tap=0.0;
	data->pitch_gain=1;
	sound_g728_init_pwf_adapter(data, data->pwf_z_coeff, data->pwf_p_coeff);
	data->pwf_z_coeff_next[0] = data->pwf_p_coeff_next[0] = 1.0;
	data->pwf_z_coeff_obsolete_p = 0;
	sound_g728_init_bsf_adapter(data, data->sf_coeff);
	data->sf_coeff_next[0] = 1.0;
	data->sf_coeff_obsolete_p = 0;
	sound_g728_init_gain_adapter(data, data->gp_coeff);
	sound_g728_init_gain_buf(data);
	data->gp_coeff_next[0] = 1.0;
	data->gp_coeff_next[1] = -1.0;
	data->gp_coeff_obsolete_p = 0;
	data->vector_end = data->thequeue;
	ZARR(data->imp_resp);
	data->imp_resp[0] = 1.0;
	sound_g728_shape_conv(data, data->imp_resp, data->shape_energy);
	transport_swstart("g728");
}

void sound_g728_stop()
{
	SAFE_FREE(data);
}

static void encode_vector(sound_g728_t *this)
{
	unsigned short ix, tmp;	/* Computed Codebook Index */
	unsigned char c;
	int b;
	int vx;    	/* Index of Recently Read Vector  */
	int lgx;	/* Logarithmic Gain Index */

	static real QMEM *vector;	/* recently read vector in the queue */

	static real zero_response[IDIM],
		weighted_speech[IDIM],
		target[IDIM],
		normtarg[IDIM],
		cb_vec[IDIM],
		pn[IDIM];
	static real	gain =1.0, scale=1.0;
	
	vector = this->vector_end - IDIM;
	if (vector < this->thequeue) vector += QSIZE;
	vx = vector-this->thequeue;
	UPDATE(this->pwf_z_coeff);	/* Copy new coeff if flag set */
	UPDATE(this->pwf_p_coeff);
	sound_g728_pwfilter2(this, vector, weighted_speech);
	UPDATE(this->sf_coeff);
	sound_g728_zresp(this,zero_response);
	sub_sig(weighted_speech, zero_response, target);
	UPDATE(this->gp_coeff);
	gain = sound_g728_predict_gain(this);
	scale = 1.0 / gain;
	sig_scale(scale, target, normtarg);
	UPDATE(this->imp_resp);
	sound_g728_trev_conv(this, this->imp_resp, normtarg, pn);
	UPDATE(this->shape_energy);
	tmp = ix = sound_g728_cb_index(this, pn);
	tmp <<= this->bits;
	tmp |= this->buf;
	b = 10+this->bits;
	while (b >= 8) {
		c = tmp;
		do_write(c);
		tmp >>= 8;
		b -= 8;
	}
	this->bits = b;
	this->buf = tmp;
	sound_g728_cb_excitation(this, ix, cb_vec);
	sig_scale(gain, cb_vec, this->qspeech+vx);
	lgx = vx/IDIM;
	sound_g728_update_gain(this, this->qspeech+vx, this->log_gains + lgx);
	sound_g728_mem_update(this, this->qspeech+vx, this->synspeech+vx);
	this->dec_end = vx+IDIM;
	if (this->dec_end >= QSIZE) this->dec_end -= QSIZE;
	NEXT_FFASE;			/* Update vector counter  */
}

/* Update the filter coeff if we are at the correct vector in the frame */
/* ffase is the vector count (1-4) within the current frame */
static void adapt_frame(sound_g728_t *this)
{
	static real
		input [NUPDATE*IDIM],
		synth [NUPDATE*IDIM],
		lg [NUPDATE];
	int gx;	/* Index for log_gains, cycle end */
	
	/* Backward syn. filter coeff update.  Occurs after full frame (before
	   first vector) but not used until the third vector of the frame */
	FFASE(1) {
		CIRCOPY(synth,this->synspeech,this->dec_end,NUPDATE*IDIM,QSIZE);
		sound_g728_bsf_adapter(this, synth, this->sf_coeff_next); /* Compute then new coeff */
	}
	
	/* Before third vector of frame */
	FFASE(3)
		this->sf_coeff_obsolete_p = 1;	/* Copy coeff computed above(2 frames later)*/
	
	/* Gain coeff update before second vector of frame */
	FFASE(2) {
		gx = this->dec_end/IDIM;
		CIRCOPY(lg, this->log_gains, gx, NUPDATE, QSIZE/IDIM);
		sound_g728_gain_adapter(this, lg, this->gp_coeff_next);
		this->gp_coeff_obsolete_p = 1;
	}
	
	FFASE(3) {
		CIRCOPY(input,this->thequeue,this->dec_end,NUPDATE*IDIM,QSIZE);
		sound_g728_pwf_adapter(this, input, this->pwf_z_coeff_next, this->pwf_p_coeff_next);
		this->pwf_z_coeff_obsolete_p = 1;
		this->pwf_p_coeff_obsolete_p = 1;
	}
	
	FFASE(3) {
		sound_g728_iresp_vcalc(this, this->sf_coeff_next, this->pwf_z_coeff_next, this->pwf_p_coeff_next, 
				       this->imp_resp_next);
		sound_g728_shape_conv(this, this->imp_resp_next, this->shape_energy_next);
		this->shape_energy_obsolete_p = 1;
		this->imp_resp_obsolete_p = 1;
	}
}

static inline void do_write(unsigned char c)
{
	static char escape[4] = {0x00,0xff,0xff,0xff};
	if (c == 0) transport_swrite(escape,4);
	else transport_swrite(&c,1);
}
