/* 
   bttvgrab 0.15.10 [2000-03-21]
   (c) 1998, 1999 by Joerg Walter <trouble@moes.pmnet.uni-oldenburg.de>
   Maintained by: Joerg Walter
   Current version at http://moes.pmnet.uni-oldenburg.de/bttvgrab/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#define TRANSPORT_INTERNAL
#include "transport.h"
#include "error.h"

/************ Prototypes ************/

void transport_net_wstart();
void transport_net_swstart();
static inline int transport_net_iwstart(char *name, int port, int *fd, void *wsig, int len);
void transport_net_wframe(int nr);
void transport_net_swframe();
void transport_net_rframe(int *nr);
void transport_net_srframe();
void transport_net_wstop();
void transport_net_swstop();
void transport_net_abort(int x);
void transport_net_rremove(int n);
void transport_net_wremove(int n);
void transport_net_rdetect();
void transport_net_srdetect();
static inline int transport_net_irdetect(char *name, int port, void *rsig, int len);
void transport_net_rstart();
void transport_net_srstart();
void transport_net_rstop();
void transport_net_srstop();
void transport_net_unframe();

/************************************/

int transport_net_lfd = -1;
int transport_net_slfd = -1;
int transport_net_unframe_num = 0;
int transport_net_unframe_p = 0;

void transport_net_srlseek(off_t offset, int whence)
{
	if (whence == SEEK_CUR && offset > 0) {
		char buffer[1024];
		while (offset > 1024) {
			transport_sread(buffer,1024);
			offset -= 1024;
		}
		transport_sread(buffer,offset);
	}
}

void transport_net_swlseek(off_t offset, int whence)
{
	if (whence == SEEK_CUR && offset > 0) {
		char buffer[1024];
		memset(buffer,0,1024);
		while (offset > 1024) {
			transport_swrite(buffer,1024);
			offset -= 1024;
		}
		transport_swrite(buffer,offset);
	}
}

void transport_net_wstart()
{
	transport_wmulti = 0;
	transport_net_lfd = transport_net_iwstart(transport_wname,31337,&transport_wfd,&transport_wsig,sizeof(transport_wsig));
}

void transport_net_swstart()
{
	transport_net_slfd = transport_net_iwstart(transport_swname,31338,&transport_swfd,&transport_swsig,sizeof(transport_swsig));
}

static inline int transport_net_iwstart(char *name, int port, int *fd, void *wsig, int len)
{
	char *portp, *remotep;
	int c, lfd;
	struct sockaddr_in net_addr;
	int addr = 0;
	
	remotep = strchr(name,'/');
	if (remotep) *remotep = 0;

	portp = strchr(name,':');
	if (portp) {
		*portp = 0;
		port = atoi(portp+1);
		ASSERT(port > 0 && port < 65536,(MSG("invalid port address: %i"),port));
	}

	ASSERT((lfd=socket(AF_INET,SOCK_STREAM,0)) >= 0,(ERRMSG("socket")));
	
	/* set up destination address */
	memset(&net_addr,0,sizeof(net_addr));
	net_addr.sin_family=AF_INET;
	if (name[0] == 0 || (name[0] == '.' && name[1] == 0)) net_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	else NOTNEG1(net_addr.sin_addr.s_addr = inet_addr(name),(MSG("invalid internet address: %s"),name));
	net_addr.sin_port=htons(port);

	c = 1;
	ASSERT(setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &c, sizeof(c)) >= 0,(ERRMSG("setsockopt")));

	/* bind to receive address */
	ASSERT(bind(lfd,(struct sockaddr *) &net_addr,sizeof(net_addr)) >= 0,(ERRMSG("bind")));
	ASSERT(listen(lfd,10) >= 0,(ERRMSG("listen")));

	if (portp) *portp = ':';
	port = 0;
	if (remotep) {
		portp = strchr(remotep+1,':');
		if (portp) {
			*portp = 0;
			port = atoi(portp+1);
			ASSERT(port >= 0 && port < 65536,(MSG("invalid port address: %i"),port));
		}
		if (remotep[1] == 0 || (remotep[1] == '.' && remotep[2] == 0)) addr = htonl(INADDR_ANY);
		else {
			NOTNEG1(net_addr.sin_addr.s_addr = inet_addr(remotep+1),(MSG("invalid internet address: %s"),remotep+1));
			addr = net_addr.sin_addr.s_addr;
		}
		if (portp) *portp = ':';
		*remotep = '/';
	}

	do {
		if (*fd >= 0) {
			SAFE_CLOSE(*fd,"unauthorized connection");
			*fd = -1;
		}
		c = sizeof(net_addr);
		ASSERT((*fd = accept(lfd,(struct sockaddr *) &net_addr, &c)) >= 0,(ERRMSG("accept")));
	} while ((addr && net_addr.sin_addr.s_addr != addr) || (port && net_addr.sin_port != port));
	ASSERT(signal(SIGPIPE,transport_net_abort) != SIG_ERR,(ERRMSG("signal")));

	ASSERT((c = send(*fd,wsig,len,0)) >= 0 ,(ERRMSG("send")));
	ASSERT(c == len,(MSG("Network error")));

//	c = 80 * 1024;
//	if (setsockopt(*fd, SOL_SOCKET, SO_SNDBUF, (void *)&c, sizeof(c)) < 0) {
//		c = 49152;
//		ASSERT(setsockopt(*fd, SOL_SOCKET, SO_SNDBUF, (void *)&c, sizeof(c)),(ERRMSG("setsockopt")));
//	}

	return lfd;
}

void transport_net_wframe(int nr)
{
	if (transport_wmulti) transport_write_int(nr);
}

void transport_net_swframe()
{
}

void transport_net_rframe(int *nr)
{
	if (transport_rmulti) {
		if (transport_net_unframe_p) {
			transport_net_unframe_p = 0;
			*nr = transport_net_unframe_num;
			return;
		}

		*nr = transport_net_unframe_num = transport_read_int();
	}
}

void transport_net_srframe()
{
}

void transport_net_wstop()
{
	if (transport_net_lfd >= 0) {
		SAFE_CLOSE(transport_net_lfd,transport_wname);
		transport_net_lfd = -1;
	}
}

void transport_net_swstop()
{
	if (transport_net_slfd >= 0) {
		SAFE_CLOSE(transport_net_slfd,transport_swname);
		transport_net_slfd = -1;
	}
}

void transport_net_abort(int x)
{
	ERROR(MSG("Network error (%i)"),x);
	END(x);
}

void transport_net_rremove(int n)
{
}

void transport_net_wremove(int n)
{
}

void transport_net_rdetect()
{
	transport_rmulti = 0;
	transport_rfd = transport_net_irdetect(transport_rname,31337,&transport_rsig,sizeof(transport_rsig));
}

void transport_net_srdetect()
{
	transport_srfd = transport_net_irdetect(transport_srname,31338,&transport_srsig,sizeof(transport_srsig));
}

static inline int transport_net_irdetect(char *name, int port, void *rsig, int len)
{
	char *portp;
	struct sockaddr_in net_addr;
	int c, fd, i;

	portp = strchr(name,':');
	if (portp) {
		*portp = 0;
		port = atoi(portp+1);
		ASSERT(port > 0 && port < 65536,(MSG("invalid port address: %i"),port));
	}
	
	ASSERT((fd=socket(AF_INET,SOCK_STREAM,0)) >= 0,(ERRMSG("socket")));
	
	memset(&net_addr,0,sizeof(net_addr));
	net_addr.sin_family=AF_INET;
	if (name[0] == 0 || (name[0] == '.' && name[1] == '/' && name[2] == 0)) {
		net_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		NOTNEG1(net_addr.sin_addr.s_addr = inet_addr(name),(MSG("invalid internet address: %s"),name));
	}
	net_addr.sin_port=htons(port);

	ASSERT(connect(fd,(struct sockaddr *) &net_addr,sizeof(net_addr)) >= 0,(ERRMSG("connect")));

	if (portp) *portp = ':';
	else sprintf(name+strlen(name),":%i",port);

	i = 0;
	do {
		ASSERT((c = recv(fd,rsig,len,0)) >= 0,(ERRMSG("recv")));
		i++;
	} while (c == 0 && i <= 10);
	ASSERT(c != 0,(MSG("Connection refused (recv)")));
	ASSERT(c == len,(MSG("network error")));
//	c = 80 * 1024;
//	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (void *)&c, sizeof(c)) < 0) {
//		c = 32768;
//		ASSERT(setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (void *)&c, sizeof(c)),(ERRMSG("setsockopt")));
//	}

	return fd;
}

void transport_net_rstart()
{
	transport_rmulti = 0;
}

void transport_net_srstart()
{
}

void transport_net_rstop()
{
}

void transport_net_srstop()
{
}

void transport_net_unframe()
{
	transport_net_unframe_p = 1;
}
